#include "TROOT.h"
#include "TStyle.h"
#include "TFile.h"
#include "TTree.h"
#include "TCanvas.h"
#include "TTimer.h"
#include "TGFrame.h"
#include "TGTab.h"
#include "TGStatusBar.h"
#include "TRootEmbeddedCanvas.h"
#include "RQ_OBJECT.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TRandom.h"
#include "TUnixSystem.h"
#include "TApplication.h"
#include "sys/file.h"
#include "sys/time.h"
#include "libgen.h"
#include "stdint.h"
#include "fcntl.h"
#include "parse.h"

#define DEBUG 0
#define ELOG 0

#if DEBUG > 0
#define PRINTD(...) printf(__VA_ARGS__); fflush(stdout);
#else
#define PRINTD(...)
#endif

#define FILENAME "/tmp/test.root"

#define TIMESTAMP_STEP_NS 1

struct Mapping
{
	const char *branch;
	const char *name;
};


/* Calibration */

#include "calib/2014/run503.h"

/* TPAT */
#define TPAT_LABR_SINGLE_L 1
#define TPAT_HPGE_SINGLE_L 2
#define TPAT_COINCIDENCE_LL 4
#define TPAT_COINCIDENCE_LH 8
#define TPAT_LABR_M 16
#define TPAT_HPGE_M 32
#define TPAT_LABR_SINGLE_H 64
#define TPAT_HPGE_SINGLE_H 128
#define TPAT_COINCIDENCE_HL 1024

//#define USED_CONDITION tpat & TPAT_LABR_SINGLE
#define USED_CONDITION 1

#define RED     "\x1b[31m"
#define GREEN   "\x1b[32m"
#define YELLOW  "\x1b[33m"
#define BLUE    "\x1b[34m"
#define MAGENTA "\x1b[35m"
#define CYAN    "\x1b[36m"
#define BLACK   "\x1b[0m"

/* Branch names */
/* QDC: LABR*EHIGH, LABR*ELOW
 * SIS3316: ADC*CH*MAXE* ADC*CH*T*
 * */
static const char* b_names[] = {
	"LABR1EHIGH", "LABR1ELOW", "LABR2EHIGH", "LABR2ELOW", // 0
	"LABR3EHIGH", "LABR3ELOW", "LABR4EHIGH", "LABR4ELOW", // 4
	"LABR5EHIGH", "LABR5ELOW", "LABR6EHIGH", "LABR6ELOW", // 8
	"LABR7EHIGH", "LABR7ELOW", "LABR8EHIGH", "LABR8ELOW", // 12
	"HPGE1E", "HPGE1B", "HPGE2E", "HPGE2B", // 16
	"HPGE3E", "HPGE3B", "HPGE4E", "HPGE4B", // 20
	"LABR1T", "LABR2T", "LABR3T", "LABR4T", // 24
	"LABR5T", "LABR6T", "LABR7T", "LABR8T", // 28
	"HPGE1T", "HPGE2T", "HPGE3T", "HPGE4T", // 32
	"MT", "TPAT", "TS_LO", "TS_HI",     // 36
	"TIMESTAMP", "", "", "",                // 40
	"LABR1S", "LABR2S", "LABR3S", "LABR4S", // 44
	"LABR5S", "LABR6S", "LABR7S", "LABR8S", // 48
	"HPGE1S", "HPGE2S", "HPGE3S", "HPGE4S", // 52
	"ZERODEG", "", "", "", // 56 
	"", "", "", "", // 60 
	"HPGEM1E","HPGEM2E","HPGEM3E","HPGEM4E", // 64 Mesytec ADC
	"HPGEM5E","HPGEM6E","HPGEM7E","HPGEM8E", // 68 Mesytec ADC
	"LABRL1S", "LABRL2S", "LABRL3S", "LABRL4S", // 72 (Labr Low Thr scalers)
	"LABRL5S", "LABRL6S", "LABRL7S", "LABRL8S", // 76 (Labr Low Thr scalers)
	"HPGEL1S", "HPGEL2S", "HPGEL3S", "HPGEL4S", // 80 (HPGE CRM scalers)
	"TRIGT1", "TRIGT2", "TRIGT3", "TRIGT4",     // 84 (Trigger times)
	"TRIGT5", "TRIGT6", "TRIGT7", "TRIGT8",     // 88 (Trigger times)
	"TRIGT9", "TRIGT10", "TRIGT11", "TRIGT12",  // 92 (Trigger times)
	"TRIGT13", "TRIGT14", "TRIGT15", "TRIGT16", // 96 (Trigger times)
	"TRIGS1", "TRIGS2", "TRIGS3", "TRIGS4",     // 100 (Trigger scalers)
	"TRIGS5", "TRIGS6", "TRIGS7", "TRIGS8",     // 104 (Trigger scalers)
	"TRIGS9", "TRIGS10", "TRIGS11", "TRIGS12",  // 108 (Trigger scalers)
	//"TRIG13S", "TRIG14S", "TRIG15S", "TRIG16S",  // 112 (Trigger scalers)
	"CRM1","CRM2","CRM3","CRM4", // 112 (HPGe Count rate meter values)
	"HPGEPILEUP1","HPGEPILEUP2","HPGEPILEUP3","HPGEPILEUP4", // 116 (Hpge pileup)
	"LABRL1T","LABRL2T","LABRL3T","LABRL4T", // Empty 120
	"HPGEL1T","HPGEL2T","HPGEL3T","HPGEL4T", // Empty 124
	"LABR1BHIGH", "LABR1BLOW", "LABR2BHIGH", "LABR2BLOW", // 128
	"LABR3BHIGH", "LABR3BLOW", "LABR4BHIGH", "LABR4BLOW", // 132
	"LABR5BHIGH", "LABR5BLOW", "LABR6BHIGH", "LABR6BLOW", // 136
	"LABR7BHIGH", "LABR7BLOW", "LABR8BHIGH", "LABR8BLOW", // 140
	"ADC1CH1MAXE1", "ADC1CH5MAXE1", "ADC1CH9MAXE1", "ADC1CH13MAXE1", // 144 /* LaBr */
	"ADC1CH1MAXE2", "ADC1CH5MAXE2", "ADC1CH9MAXE2", "ADC1CH13MAXE2", // 148
	"ADC3CH1MAXE1", "ADC3CH5MAXE1", "ADC3CH9MAXE1", "ADC3CH13MAXE1", // 152 /* Germanium */
	"ADC3CH1MAXE2", "ADC3CH5MAXE2", "ADC3CH9MAXE2", "ADC3CH13MAXE2", // 156
};

static const char* trlo_names[] = {
	"TRLORAW1","TRLORAW2","TRLORAW3","TRLORAW4",
	"TRLORAW5","TRLORAW6","TRLORAW7","TRLORAW8",
	"TRLORAW9","TRLORAW10","TRLORAW11","TRLORAW12",
	"TRLORAW13","TRLORAW14","TRLORAW15","TRLORAW16",
	"TRLOBDT1","TRLOBDT2","TRLOBDT3","TRLOBDT4",
	"TRLOBDT5","TRLOBDT6","TRLOBDT7","TRLOBDT8",
	"TRLOBDT9","TRLOBDT10","TRLOBDT11","TRLOBDT12",
	"TRLOBDT13","TRLOBDT14","TRLOBDT15","TRLOBDT16",
	"TRLOADT1","TRLOADT2","TRLOADT3","TRLOADT4",
	"TRLOADT5","TRLOADT6","TRLOADT7","TRLOADT8",
	"TRLOADT9","TRLOADT10","TRLOADT11","TRLOADT12",
	"TRLOADT13","TRLOADT14","TRLOADT15","TRLOADT16",
	"TRLOARD1","TRLOARD2","TRLOARD3","TRLOARD4",
	"TRLOARD5","TRLOARD6","TRLOARD7","TRLOARD8",
	"TRLOARD9","TRLOARD10","TRLOARD11","TRLOARD12",
	"TRLOARD13","TRLOARD14","TRLOARD15","TRLOARD16"
};

static const char* trlo_raw_names[] = {
	"LaBrOR L", "LaBrOR H", "HPGeOR L", "HPGeOR H",
	"LaBr M L", "LaBr M H", "HPGe M L", "HPGe M H",
	"  ---   ", "HPGe0deg", "RF      ", "Pulser  ",
	"  ---   ", "  ---   ", "  ---   ", "  ---   "
};

static const char* trlo_tpat_names[] = {
	"Singl LaBr H ", "Singl HPGe H ", "Coinc. L-L   ", "Coinc  L-H   ",
	"LaBr M H     ", "HPGe M H     ", "Zero Degree  ", "Pulser       ",
	"Singl LaBr  L", "Singl HPGe  L", "Coinc  H-L   ", "-            ",
	"-            ", "-            ", "-            ", "-            "
};

typedef struct trlo_scaler_s {
	uint32_t raw;
	uint32_t bdt;
	uint32_t adt;
	uint32_t ard;
} trlo_scaler;

typedef struct trlo_scaler64_s {
	uint64_t raw;
	uint64_t bdt;
	uint64_t adt;
	uint64_t ard;
} trlo_scaler64;


static const char* bc_names[] = {
	"LABR1EC","LABR2EC","LABR3EC","LABR4EC",
	"LABR5EC","LABR6EC","LABR7EC","LABR8EC", // 0
	"HPGE1EC","HPGE2EC","HPGE3EC","HPGE4EC",
	"HPGEM1EC","HPGEM2EC","HPGEM3EC","HPGEM4EC", // 8
	"LABR1ELOWC","LABR2ELOWC","LABR3ELOWC","LABR4ELOWC",
	"LABR5ELOWC","LABR6ELOWC","LABR7ELOWC","LABR8ELOWC" // 16
};

static const char* bcr_names[] = {
	"LABR1ECR","LABR2ECR","LABR3ECR","LABR4ECR",
	"LABR5ECR","LABR6ECR","LABR7ECR","LABR8ECR", // 0
	"HPGE1ECR","HPGE2ECR","HPGE3ECR","HPGE4ECR",
	"HPGEM1ECR","HPGEM2ECR","HPGEM3ECR","HPGEM4ECR", // 8
};

/* Histogram names */
static const char* h_names[] = {
	"Labr1",  "Labr2",  "Labr3",  "Labr4",                // 0
	"Labr1_low",  "Labr2_low",  "Labr3_low",  "Labr4_low",// 4
	"HPGe1",  "HPGe2",  "HPGe3",  "HPGe4",                // 8
	"Labr1_t vs. RF", "Labr2_t vs. RF", "Labr3_t vs. RF", "Labr4_t vs. RF",  // 12
	"Labr5_t vs. RF", "Labr6_t vs. RF", "Labr7_t vs. RF", "Labr8_t vs. RF",  // 16
	"HPGe1_t",  "HPGe2_t",  "HPGe3_t",  "HPGe4_t",                           // 20
	"LaBr Sum", "HPGe Sum", "Coincidence LaBr-LaBr", "Coincidence HPGe-LaBr",// 24
	"LaBr Mult", "HPGe Mult", "Multiplicity", "Rates (calculated)",          // 28
	"Scaler LaBr", "Scaler HPGe", "Scaler Trigger", "",               // 32
	"HPGeMesy1",  "HPGeMesy2",  "HPGeMesy3",  "HPGeMesy4",            // 36
	"HPGeMesy5",  "", "", "",                                         // 40
	"Labr1_t_raw",  "Labr2_t_raw",  "Labr3_raw",  "Labr4_raw",        // 44
	"Labr5_t_raw",  "Labr6_t_raw",  "Labr7_raw",  "Labr8_t_raw",      // 48
	"HPGe1_t_raw",  "HPGe2_t_raw",  "HPGe3_raw",  "HPGe4_raw",        // 52
	"LabrLow1",  "LabrLow2",  "LabrLow3",  "LabrLow4",                // 56
	"LabrLow5",  "LabrLow6",  "LabrLow7",  "LabrLow8",                // 60
	"Labr1_t_low",  "Labr2_t_low",  "Labr3_low",  "Labr4_low",        // 64
	"HPGe1_t_low",  "HPGe2_t_low",  "HPGe3_low",  "HPGe4_low",        // 68
};


static const int h_ltr = 48;
static const int h_htr = 56;

static const char* hc_names[] = {
	"Labr1_c",  "Labr2_c",  "Labr3_c",  "Labr4_c",                // 0
	"Labr5_c",  "Labr6_c",  "Labr7_c",  "Labr8_c",                // 4
	"HPGe1_c",  "HPGe2_c",  "HPGe3_c",  "HPGe4_c",                // 8
	"Labr1_t",  "Labr2_t",  "Labr3_t",  "Labr4_t",        // 12
	"Labr5_t",  "Labr6_t",  "Labr7_t",  "Labr8_t",        // 16
	"HPGe1_t",  "HPGe2_t",  "HPGe3_t",  "HPGe4_t",        // 20
	"LaBr Sum", "HPGe Sum", "",  "Coincidence Matrix",    // 24
	"LaBr Mult", "HPGe Mult", "Multiplicity", "Rates (calculated)", // 28
	"Scaler LaBr", "Scaler HPGe", "Scaler Trigger", "",   // 32
	"Labr1 condition 1", "Labr2 condition 1", 
	"Labr3 condition 1", "Labr4 condition 1", // 36
	"Labr5 condition 1", "Labr6 condition 1", 
	"Labr7 condition 1", "Labr8 condition 1", // 40
	"Labr1 condition 2", "Labr2 condition 2", 
	"Labr3 condition 2", "Labr4 condition 2", // 44
	"Labr5 condition 2", "Labr6 condition 2", 
	"Labr7 condition 2", "Labr8 condition 2"  // 48
};

static const char* h2_names[] = {
	"","","","",
	"HPGe1 vs. Time", "HPGe2 vs. Time", 
	"HPGe3 vs. Time", "HPGe4 vs. Time",    // 4
	"HPGe1 over Time", "HPGe2 over Time", 
	"HPGe3 over Time", "HPGe4 over Time" // 8
};

/* Histogram ranges */

/*static const double h_min[] = {
  0, 0, 0, 0,                   // 0
  0, 0, 0, 0,                   // 4
  0, 0, 0, 0,                   // 8
  -250, -250, -250, -250,       // 12
  -250, -250, -250, -250,       // 16
  -500, -500, -500, -500,   // 20
  0, 0, -1000, 0,               // 24
  -0.5, -0.5, -0.5, -0.5,       // 28
  0.5, 0.5, 0.5, 0,             // 32
  0, 0, 0, 0,                   // 36
  0, 0, 0, 0,                   // 40
  0, 0, 0, 0,                   // 44
  0, 0, 0, 0,                   // 48
  0, 0, 0, 0                    // 52
  };*/

static const double h_min[] = {
	0, 0, 0, 0,                   // 0
	0, 0, 0, 0,                   // 4
	0, 0, 0, 0,                   // 8
	-500, -500, -500, -500,       // 12
	-500, -500, -500, -500,       // 16
	-500, -500, -500, -500,   // 20
	0, 0, -1000, 0,               // 24
	-0.5, -0.5, -0.5, -0.5,       // 28
	0.5, 0.5, 0.5, 0,             // 32
	0, 0, 0, 0,                   // 36
	0, 0, 0, 0,                   // 40
	-500, -500, -500, -500,       // 44
	-500, -500, -500, -500,       // 48
	-500, -500, -500, -500,       // 52
	0, 0, 0, 0, // 56
	0, 0, 0, 0,  // 60
	-500, -500, -500, -500,  // 64
	-500, -500, -500, -500  // 68
};


static const double h_max[] = {
	80000, 80000, 80000, 80000,
	80000, 80000, 80000, 80000,
	9000000, 6000000, 6000000, 6000000,
	500, 500, 500, 500,
	500, 500, 500, 500,
	500, 500, 500, 500,
	4000, 10000, 1000, 1000,
	10.5, 10.5, 10.5, 0.5,
	8.5, 4.5, 4.5, 0,
	8191, 8191, 8191, 8191,
	8191, 0, 0, 0,
	500, 500, 500, 500,
	500, 500, 500, 500,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000
};

static const double h_bins[] = {
	8001, 8001, 8001, 8001,
	8001, 8001, 8001, 8001,
	16001, 16001, 16001, 16001,
	8001, 8001, 8001, 8001,
	8001, 8001, 8001, 8001,
	4001, 4001, 4001, 4001,
	4001, 10001, 1001, 1001,
	11, 11, 11, 0,
	8, 4, 4, 0,
	8192, 8192, 8192, 8192, 
	8192, 0, 0, 0,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001,
	4001, 4001, 4001, 4001
};

static const double h_min_c[] = {
	0, 0, 0, 0,                   // 0
	0, 0, 0, 0,                   // 4
	0, 0, 0, 0,                   // 8
	-1000, -1000, -1000, -1000,   // 12
	-1000, -1000, -1000, -1000,   // 16
	-1000, -1000, -1000, -1000,   // 20
	0, 0, -1000, 0,               // 24
	-0.5, -0.5, -0.5, -0.5,       // 28
	0.5, 0.5, 0.5, 0,             // 32
	0, 0, 0, 0,                   // 36
	0, 0, 0, 0,                   // 40
	0, 0, 0, 0,                   // 44
	0, 0, 0, 0                    // 48
};

static const double h_max_c[] = {
	12000, 12000, 12000, 12000,
	10000, 10000, 10000, 10000,
	12000, 12000, 12000, 12000,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000,
	4000, 4000, 4000, 4000,
	12000, 12000, 1000, 1000,
	10.5, 10.5, 10.5, 0.5,
	8.5, 4.5, 4.5, 0,
	10000, 10000, 10000, 10000,
	4000, 4000, 4000, 4000,
	10000, 10000, 10000, 10000,
	4000, 4000, 4000, 4000
};

/*
   static const double h_max_c[] = {
   6200, 6200, 6200, 6200,
   6200, 6200, 6200, 6200,
   12000, 12000, 12000, 12000,
   4000, 4000, 4000, 4000,
   4000, 4000, 4000, 4000,
   4000, 4000, 4000, 4000,
   12000, 10000, 1000, 1000,
   10.5, 10.5, 10.5, 0.5,
   8.5, 4.5, 4.5, 0,
   10000, 10000, 10000, 10000,
   4000, 4000, 4000, 4000,
   10000, 10000, 10000, 10000,
   4000, 4000, 4000, 4000
   };
   */

/* This is given in pairs of two x1,y1,x2,y2,... */
static const int h2_bins[] = {
	0,0, 0,0, 0,0, 0,0,
	1000,500, 1000,500, 1000,500, 1000,500,
	240,8000, 240,8000, 240,8000, 240,8000
};

static const double h2_min[] = {
	0,-500, 0,-500, 0,-500, 0,-500,
	0,-0, 0,0, 0,0, 0,0,              // HPGE E vs. ToF
	0,0, 0,0, 0,0, 0,0, 0,0
};

static const double h2_max[] = {
	4000,500, 4000,500, 4000,500, 4000,500,
	4000,400, 4000,400, 4000,400, 4000,400,     // HPGE E vs. ToF
	1e5,16000, 1e5,16000, 1e5,16000, 1e5,16000
};


class myTH1F : public TH1F {
	public:
		myTH1F(const char* name, const char* title, Int_t nbinsx, Double_t xlow, Double_t xup) 
			: TH1F(name, title, nbinsx, xlow, xup) {}

		void ResetHist() {this->Reset();} // *MENU*
		ClassDef(myTH1F,1);
};

void onKeyPress() {
	if (gPad->GetEvent() == kKeyPress) {
		int key = gPad->GetEventX();
		if (key == 'r') {
			((myTH1F*)(gPad->GetListOfPrimitives()->At(1)))->ResetHist();
		}
	}
}

class TreeReader : public TQObject {

	RQ_OBJECT("TreeReader");

	public:

	TreeReader(char* _filename, char* _options = 0) {
		filename = _filename;
		TFile* file = new TFile(filename,"READ");
		f = file;
		ntuple = (TTree*)f->Get("h101");

		memset(trlo_sc,0,sizeof(trlo_sc));
		memset(trlo_sc_sum,0,sizeof(trlo_sc_sum));
		memset(trlo_sc_start,0,sizeof(trlo_sc_start));


		setBranchAddresses();
		createHistograms();

		first = 0; 

		if(_options == 0 || strlen(_options) == 0) {
			struct timeval now, last_draw_time;

			gettimeofday(&last_draw_time, NULL);
			setupGui();
			styleHistograms();
			setLogScale();

			//Connect("StatusBarChanged()","TreeReader",this,"RedrawStatusBar()");

			t1 = new TTimer();
			//t1->Connect("Timeout()","TreeReader",this,"draw()");
			t1->Start(500);
			t2 = new TTimer();
			//t2->Connect("Timeout()","TreeReader",this,"RedrawStatusBar()");
			t2->Start(500);

			while(!gSystem->ProcessEvents()) {
				gettimeofday(&now, NULL);
				if (now.tv_sec > last_draw_time.tv_sec) {
					draw();
					last_draw_time = now;
				}
				/*exit(1);*/
			}
			gSystem->Abort();
		} else {
			int showProgress = 1;
			int showScalers = 0;
			if(strncmp(_options,"tv",2) == 0) {
				printf("Creating tv spectra\n");
				int only_tv = 1;
				printRealTime();
				fillHistograms(showProgress, only_tv);
				writeTv();
			} else if(strncmp(_options,"stat",4) == 0) {
				printf("Statistics\n");
				showScaler(showProgress,showScalers);
			} else if(strncmp(_options,"recal",5) == 0) {
				int n_peaks;
				int ret = sscanf(_options+5,"%d",&n_peaks);
				if (ret != 1) {
					printf("Must specify number of peaks in 'recal' mode. %s (e.g.: recal n)\n",_options);
					exit(1);
				}
				printf("Recalibration with %d peaks\n",n_peaks);
				recalibrateTree(showProgress,n_peaks);
			} else {
				if(strncmp(_options,"cal",3) != 0) {
					printf("Unknown option %s\n",_options);
					return;
				}
				extendTree(showProgress);
			}
			gSystem->Abort();
		}
	}

	~TreeReader() {
		f->Close();
		delete f;
		f = 0;
		frame->Cleanup();
		delete frame;
	} 



	void createHistograms() {

		memset(h,0,sizeof(h));
		memset(hc,0,sizeof(h));

		TStyle* myStyle = gROOT->GetStyle("Modern");
		myStyle->SetOptStat("neimr");
		myStyle->SetStatH(0.3);
		myStyle->SetStatW(0.2);
		myStyle->cd();

		for(int i = 0; i < 8; ++i) {
			// Labr energy spectra
			h[i] = new myTH1F(h_names[i],h_names[i],h_bins[i],h_min[i],h_max[i]);
			hc[i] = new myTH1F(hc_names[i],hc_names[i],h_bins[i],h_min_c[i],h_max_c[i]);
			// LaBr energy spectra with applied condition
			hc[i+16] = new myTH1F(hc_names[i+36],hc_names[i+36],h_bins[i+36],h_min_c[i+36],h_max_c[i+36]);
			hc[i+24] = new myTH1F(hc_names[i+44],hc_names[i+44],h_bins[i+44],h_min_c[i+44],h_max_c[i+44]);
			// Labr time spectra
			h[12+i] = new myTH1F(h_names[12+i],h_names[12+i],h_bins[12+i],h_min[12+i],h_max[12+i]);
			h[48+i] = new myTH1F(h_names[44+i],h_names[44+i],h_bins[44+i],h_min[44+i],h_max[44+i]);
			//if(i < 4)
			//  h[64+i] = new myTH1F(h_names[44+i],h_names[44+i],h_bins[44+i],h_min[44+i],h_max[44+i]);
			if(i < 4) {
				// HPGe energy spectra
				h[8+i] = new myTH1F(h_names[8+i],h_names[8+i],h_bins[8+i],h_min[8+i],h_max[8+i]);
				hc[8+i] = new myTH1F(hc_names[8+i],hc_names[8+i],h_bins[8+i],h_min_c[8+i],h_max_c[8+i]);
				// HPGe time spectra
				h[20+i] = new myTH1F(h_names[20+i],h_names[20+i],h_bins[20+i],h_min[20+i],h_max[20+i]);
				h[48+8+i] = new myTH1F(h_names[52+i],h_names[52+i],h_bins[52+i],h_min[52+i],h_max[52+i]);
				// HPGe time-energy spectra
				h2[4+i] = new TH2F(h2_names[4+i],h2_names[4+i],
						h2_bins[((4+i)*2)+1],h2_min[((4+i)*2)+1],h2_max[((4+i)*2)+1],
						h2_bins[((4+i)*2)],h2_min[(4+i)*2],h2_max[(4+i)*2]
						);
				// HPGe energy spectra over time (to check for drifts)
				h2[8+i] = new TH2F(h2_names[8+i],h2_names[8+i],
						h2_bins[(8+i)*2],h2_min[(8+i)*2],h2_max[(8+i)*2],
						h2_bins[(8+i)*2+1],h2_min[(8+i)*2+1],h2_max[(8+i)*2+1]
						);
			}
		}
		/* Special spectra */
		// Sum spectra
		h[32+0] = new myTH1F(h_names[24+0], h_names[24+0], h_bins[24+0],h_min[24+0],h_max_c[24+0]);
		h[32+1] = new myTH1F(h_names[24+1], h_names[24+1], h_bins[24+1],h_min[24+1],h_max_c[24+1]);
		// Coincidence Times
		// h[32+2] = new TH1F(h_names[24+2], h_names[24+2], h_bins[24+2],h_min[24+2],h_max[24+2]);
		// Multiplicity
		h[35+0] = new myTH1F(h_names[28+0], h_names[28+0], h_bins[28+0],h_min[28+0],h_max[28+0]);
		h[35+1] = new myTH1F(h_names[28+1], h_names[28+1], h_bins[28+1],h_min[28+1],h_max[28+1]);
		h[35+2] = new myTH1F(h_names[28+2], h_names[28+2], h_bins[28+2],h_min[28+2],h_max[28+2]);
		// Detector rates
		h[38]   = new myTH1F(h_names[28+3], h_names[28+3], 12, 1, 12);

		// Scalers
		for(int i = 0; i < 3; ++i) {
			h[39+i] = new myTH1F(h_names[32+i], h_names[32+i], h_bins[32+i],h_min[32+i],h_max[32+i]);
			h[39+i]->GetYaxis()->SetRangeUser(0,100000);
		}

		// HPGe from Mesytec ADC
		for(int i = 0; i < 5; ++i) {
			h[42+i] = new myTH1F(h_names[36+i], h_names[36+i], h_bins[36+i],h_min[36+i],h_max[36+i]);

		}

		// Coincidence
		//    h2[0] = new TH2F(h_names[24+3], h_names[24+3], 1000,0,6200, 6201,0,6200);
		h2[0] = new TH2F(h_names[24+3], h_names[24+3], 1000,0,10000, 5000,0,10000);
		//    h2[2] = new TH2F(h_names[24+2], h_names[24+2], 1000,0,6200, 1000,0,6200);
		h2[2] = new TH2F(h_names[24+2], h_names[24+2], 1000,0,10000, 1000,0,10000);

		// Correlation
		h2[1] = new TH2F("Correlation", "Correlation", 12,0.5,12.5, 12,0.5,12.5);

	}

	void setLogScale() {
		TVirtualPad* vp;
		for(int i = 0; i < 8; ++i) {
			vp = ec_l->GetCanvas()->cd(i+1);
#if ELOG
			vp->SetLogy();
#endif
			vp = ec_5->GetCanvas()->cd(i+1);
			vp->SetLogy();
			vp = ec_8->GetCanvas()->cd(i+1);
			//vp->SetLogy(); // 233U

			if(i < 4) {
				vp = ec_h->GetCanvas()->cd(i+1);
#if ELOG
				vp->SetLogy();
#endif
				vp = ec_h->GetCanvas()->cd(4+i+1);
				//vp->SetLogy(); // 233U
				vp = ec_8->GetCanvas()->cd(8+i+1);
				// vp->SetLogy(); // 233U
			}
		}
	}

	void styleHistograms() {
		for(int i = 0; i < 64; ++i) {
			if(h[i] != 0) {
				h[i]->GetXaxis()->SetLabelSize(0.06);
				h[i]->GetYaxis()->SetLabelSize(0.05);
			}
			if(hc[i] != 0) {
				hc[i]->GetXaxis()->SetLabelSize(0.06);
				hc[i]->GetYaxis()->SetLabelSize(0.05);
				//hc[i]->GetXaxis()->SetRangeUser(2500,3500); // 233U
			}
		}
		h2[0]->GetXaxis()->SetLabelSize(0.04);
		h2[0]->GetYaxis()->SetLabelSize(0.04);
		h2[2]->GetXaxis()->SetLabelSize(0.04);
		h2[2]->GetYaxis()->SetLabelSize(0.04);
		h[32+0]->GetXaxis()->SetLabelSize(0.04);
		h[32+0]->GetYaxis()->SetLabelSize(0.04);
		h[32+1]->GetXaxis()->SetLabelSize(0.04);
		h[32+1]->GetYaxis()->SetLabelSize(0.04);
		//    for(int i = 35; i < 42; ++i) {
		//      if(h[i] != 0) {
		//      }
		//    }
	}

	void setupGui() {
		frame = new TGMainFrame(gClient->GetRoot(),400,400);

		tab = new TGTab(frame,1200,900);
		labr_frame = tab->AddTab("LaBr E");
		frame_5    = tab->AddTab("LaBr T");
		frame_6    = tab->AddTab("LaBr E calib");
		frame_9    = tab->AddTab("LaBr w/Cond");
		hpge_frame = tab->AddTab("HPGe");
		frame_7    = tab->AddTab("HPGe calib");
		frame_3    = tab->AddTab("Sum/Coinc");
		frame_4    = tab->AddTab("Multiplicity/Scalers");
		frame_8    = tab->AddTab("RAW Times");
		cf[9] = tab->AddTab("Timelines");

		status = new TGStatusBar(frame,50,10,kHorizontalFrame);
		status->SetParts(3);

		ec_l = new TRootEmbeddedCanvas("canvas_labr",labr_frame,1200,900);
		ec_l->GetCanvas()->Divide(2,2);
		ec_l->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_l->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_l->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_l->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_l->GetCanvas()->cd(1)->AddExec("ec0_1_exec","onKeyPress()");
		ec_l->GetCanvas()->cd(2)->AddExec("ec0_2_exec","onKeyPress()");
		ec_l->GetCanvas()->cd(3)->AddExec("ec0_3_exec","onKeyPress()");
		ec_l->GetCanvas()->cd(4)->AddExec("ec0_4_exec","onKeyPress()");
		ec_h = new TRootEmbeddedCanvas("canvas_hpge",hpge_frame,1200,900);
		ec_h->GetCanvas()->Divide(2,4);
		ec_h->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(5)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(6)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(7)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(8)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(1)->AddExec("ec1_1_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(2)->AddExec("ec1_2_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(3)->AddExec("ec1_3_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(4)->AddExec("ec1_4_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(5)->AddExec("ec1_5_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(6)->AddExec("ec1_6_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(7)->AddExec("ec1_7_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(8)->AddExec("ec1_8_exec","onKeyPress()");
		ec_3 = new TRootEmbeddedCanvas("canvas_3",frame_3,1200,900);
		ec_3->GetCanvas()->Divide(2,2);
		ec_3->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_3->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_3->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_3->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_3->GetCanvas()->cd(1)->AddExec("ec3_1_exec","onKeyPress()");
		ec_3->GetCanvas()->cd(2)->AddExec("ec3_2_exec","onKeyPress()");
		ec_3->GetCanvas()->cd(3)->AddExec("ec3_3_exec","onKeyPress()");
		ec_3->GetCanvas()->cd(4)->AddExec("ec3_4_exec","onKeyPress()");
		ec_4 = new TRootEmbeddedCanvas("canvas_4",frame_4,1200,900);
		ec_4->GetCanvas()->Divide(4,2);
		ec_4->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(5)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(6)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(7)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(8)->SetBit(TPad::kCannotMove);
		ec_4->GetCanvas()->cd(1)->AddExec("ec4_1_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(2)->AddExec("ec4_2_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(3)->AddExec("ec4_3_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(4)->AddExec("ec4_4_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(5)->AddExec("ec4_5_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(6)->AddExec("ec4_6_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(7)->AddExec("ec4_7_exec","onKeyPress()");
		ec_4->GetCanvas()->cd(8)->AddExec("ec4_8_exec","onKeyPress()");
		ec_5 = new TRootEmbeddedCanvas("canvas_labr_t",frame_5,1200,900);
		ec_5->GetCanvas()->Divide(2,2);
		ec_5->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_5->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_5->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_5->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_5->GetCanvas()->cd(1)->AddExec("ec5_1_exec","onKeyPress()");
		ec_5->GetCanvas()->cd(2)->AddExec("ec5_2_exec","onKeyPress()");
		ec_5->GetCanvas()->cd(3)->AddExec("ec5_3_exec","onKeyPress()");
		ec_5->GetCanvas()->cd(4)->AddExec("ec5_4_exec","onKeyPress()");
		ec_6 = new TRootEmbeddedCanvas("labr_e_c",frame_6,1200,900);
		ec_6->GetCanvas()->Divide(2,2);
		ec_6->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_6->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_6->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_6->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_6->GetCanvas()->cd(1)->AddExec("ec6_1_exec","onKeyPress()");
		ec_6->GetCanvas()->cd(2)->AddExec("ec6_2_exec","onKeyPress()");
		ec_6->GetCanvas()->cd(3)->AddExec("ec6_3_exec","onKeyPress()");
		ec_6->GetCanvas()->cd(4)->AddExec("ec6_4_exec","onKeyPress()");
		ec_7 = new TRootEmbeddedCanvas("hpge_e_c",frame_7,1200,900);
		ec_7->GetCanvas()->Divide(2,2);
		ec_7->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_7->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_7->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_7->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_7->GetCanvas()->cd(1)->AddExec("ec7_1_exec","onKeyPress()");
		ec_7->GetCanvas()->cd(2)->AddExec("ec7_2_exec","onKeyPress()");
		ec_7->GetCanvas()->cd(3)->AddExec("ec7_3_exec","onKeyPress()");
		ec_7->GetCanvas()->cd(4)->AddExec("ec7_4_exec","onKeyPress()");
		ec_8 = new TRootEmbeddedCanvas("raw_times",frame_8,1200,900);
		ec_8->GetCanvas()->Divide(2,4);
		ec_8->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(5)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(6)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(7)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(8)->SetBit(TPad::kCannotMove);
		ec_8->GetCanvas()->cd(1)->AddExec("ec8_1_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(2)->AddExec("ec8_2_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(3)->AddExec("ec8_3_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(4)->AddExec("ec8_4_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(5)->AddExec("ec8_5_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(6)->AddExec("ec8_6_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(7)->AddExec("ec8_7_exec","onKeyPress()");
		ec_8->GetCanvas()->cd(8)->AddExec("ec8_8_exec","onKeyPress()");
		ec_9 = new TRootEmbeddedCanvas("labr_cond",frame_9,1200,900);
		ec_9->GetCanvas()->Divide(2,4);
		ec_9->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(5)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(6)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(7)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(8)->SetBit(TPad::kCannotMove);
		ec_9->GetCanvas()->cd(1)->AddExec("ec9_1_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(2)->AddExec("ec9_2_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(3)->AddExec("ec9_3_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(4)->AddExec("ec9_4_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(5)->AddExec("ec9_5_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(6)->AddExec("ec9_6_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(7)->AddExec("ec9_7_exec","onKeyPress()");
		ec_9->GetCanvas()->cd(8)->AddExec("ec9_8_exec","onKeyPress()");
		ec[9] = new TRootEmbeddedCanvas("timelines",cf[9],1200,900);
		ec[9]->GetCanvas()->Divide(2,2);
		ec[9]->GetCanvas()->cd(1)->SetBit(TPad::kCannotMove);
		ec[9]->GetCanvas()->cd(2)->SetBit(TPad::kCannotMove);
		ec[9]->GetCanvas()->cd(3)->SetBit(TPad::kCannotMove);
		ec[9]->GetCanvas()->cd(4)->SetBit(TPad::kCannotMove);
		ec_h->GetCanvas()->cd(1)->AddExec("ec1_1_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(2)->AddExec("ec1_2_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(3)->AddExec("ec1_3_exec","onKeyPress()");
		ec_h->GetCanvas()->cd(4)->AddExec("ec1_4_exec","onKeyPress()");

		labr_frame->AddFrame(ec_l, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		labr_frame->Resize(labr_frame->GetDefaultSize());

		hpge_frame->AddFrame(ec_h, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		hpge_frame->Resize(hpge_frame->GetDefaultSize());

		frame_3->AddFrame(ec_3, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_3->Resize(frame_3->GetDefaultSize());

		frame_4->AddFrame(ec_4, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_4->Resize(frame_4->GetDefaultSize());

		frame_5->AddFrame(ec_5, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_5->Resize(frame_5->GetDefaultSize());

		frame_6->AddFrame(ec_6, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_6->Resize(frame_6->GetDefaultSize());

		frame_7->AddFrame(ec_7, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_7->Resize(frame_7->GetDefaultSize());

		frame_8->AddFrame(ec_8, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_8->Resize(frame_8->GetDefaultSize());

		frame_9->AddFrame(ec_9, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		frame_9->Resize(frame_9->GetDefaultSize());

		cf[9]->AddFrame(ec[9], new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
			10, 10, 10, 1));
		cf[9]->Resize(cf[9]->GetDefaultSize());

		frame->AddFrame(tab, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 
		    10, 10, 10, 1));
		frame->AddFrame(status,
		    new TGLayoutHints(kLHintsBottom | kLHintsLeft |  kLHintsExpandX,0,0,2,0)); 
		frame->MapSubwindows();
		frame->Resize(frame->GetDefaultSize());
		frame->MapWindow();

		frame->SetWindowName("GHOST - Gamma3 Histogramming and Online Spectra Tool");
	}

	void setBranchAddresses() {
		for(int i = 0; i < 4; ++i) {
			/* LABR */
			b_labr_h_e[i] = 0;
			b_labr_l_e[i] = 0;
			b_labr_t[i] = 0;
			b_labr_sis_e[i] = 0;
			b_labr_sis_b[i] = 0;
			b_labr_s[i] = 0;
			labr_s_prev[i] = 0;
			ntuple->SetBranchAddress(b_names[24+i], &labr_t[i], &b_labr_t[i]);
			ntuple->SetBranchAddress(b_names[120+i], &labrl_t[i], &b_labrl_t[i]);
			ntuple->SetBranchAddress(b_names[144+i], &labr_sis_b[i], &b_labr_sis_b[i]);
			ntuple->SetBranchAddress(b_names[148+i], &labr_sis_e[i], &b_labr_sis_e[i]);
			ntuple->SetBranchAddress(b_names[44+i], &labr_s[i], &b_labr_s[i]);
			ntuple->SetBranchAddress(b_names[72+i], &labrl_s[i], &b_labrl_s[i]);
			/*ntuple->SetBranchAddress(b_names[i*2],   &labr_h_e[i], &b_labr_h_e[i]);
			  ntuple->SetBranchAddress(b_names[i*2+1], &labr_l_e[i], &b_labr_l_e[i]);*/
			/* HPGE */
			b_hpge_e[i] = 0;
			b_hpge_b[i] = 0;
			b_hpge_t[i] = 0;
			b_hpge_s[i] = 0;
			b_hpge_pileup[i] = 0;
			b_hpge_crm[i] = 0;
			hpge_s_prev[i] = 0;
			hpge_crm_prev[i] = 0;
			ntuple->SetBranchAddress(b_names[156+i], &hpge_e[i], &b_hpge_e[i]);
			ntuple->SetBranchAddress(b_names[152+i], &hpge_b[i], &b_hpge_b[i]);
			ntuple->SetBranchAddress(b_names[32+i], &hpge_t[i], &b_hpge_t[i]);
			ntuple->SetBranchAddress(b_names[124+i], &hpgel_t[i], &b_hpgel_t[i]);
			ntuple->SetBranchAddress(b_names[52+i], &hpge_s[i], &b_hpge_s[i]);
			ntuple->SetBranchAddress(b_names[80+i], &hpgel_s[i], &b_hpgel_s[i]);
			ntuple->SetBranchAddress(b_names[112+i], &hpge_crm[i], &b_hpge_crm[i]);
			/*ntuple->SetBranchAddress(b_names[116+i], &hpge_pileup[i], &b_hpge_pileup[i]);*/
		}
		for(int i = 0; i < 12; ++i) {
			ntuple->SetBranchAddress(b_names[84+i], &trig_t[i], &b_trig_t[i]);
			ntuple->SetBranchAddress(b_names[100+i], &trig_s[i], &b_trig_s[i]);
		}
		ntuple->SetBranchAddress("TPAT", &tpat, &b_tpat);
		ntuple->SetBranchAddress("TS_LO", &time_lo, &b_time_lo);
		ntuple->SetBranchAddress("TS_HI", &time_hi, &b_time_hi);

		for(int i = 0; i < 16; ++i) {
			b_trlo_sc[i] = 0;
			b_trlo_sc[i+16] = 0;
			b_trlo_sc[i+32] = 0;
			b_trlo_sc[i+48] = 0;
			ntuple->SetBranchAddress(trlo_names[i], &trlo_sc[i].raw, &b_trlo_sc[i]);
			ntuple->SetBranchAddress(trlo_names[i+16], &trlo_sc[i].bdt, &b_trlo_sc[i+16]);
			ntuple->SetBranchAddress(trlo_names[i+32], &trlo_sc[i].adt, &b_trlo_sc[i+32]);
			ntuple->SetBranchAddress(trlo_names[i+48], &trlo_sc[i].ard, &b_trlo_sc[i+48]);
		}
	}

	int checkCuts() {
		return 0;
	}

#ifdef USE_STEPWISE_CALIB
	void calibrate_stepwise(int n) {
		for(int i = 0; i < 8; i++) {
			double labr_h = (double)(labr_h_e[i]) + r.Rndm() - 0.5;
			double labr_sis = (double)(labr_sis_e[i]) + r.Rndm() - 0.5;
			double labr_l = (double)(labr_l_e[i]) + r.Rndm() - 0.5;
			double hpge   = (double)(hpge_e[i] - hpge_b[i]) + r.Rndm() - 0.5;
			double hpge_bin = (double)(h_bins[8+i])/(double)(h_max[8+i]);

			/* High range LaBr */
			int idx = i+(n*7);
			labr_h_e_calib[i] = (labr_h * labr_h * labr_h * cubic_labr_stepwise[idx]) + 
				(labr_h * labr_h * quadratic_labr_stepwise[idx]) + 
				(labr_h * slopes_labr_stepwise[idx]) + 
				offsets_labr_stepwise[idx];
			labr_sis_e_calib[i] = (labr_sis * labr_sis * labr_sis * cubic_labr_stepwise[idx]) + 
				(labr_sis * labr_sis * quadratic_labr_stepwise[idx]) + 
				(labr_sis * slopes_labr_stepwise[idx]) + 
				offsets_labr_stepwise[idx];

			/* Low range LaBr */
			labr_l_e_calib[i] = labr_l * slopes_labr_l[i] + offsets_labr_l[i];

			if(i < 4) {
				/* Struck HPGe */
				hpge_e_calib[i] = hpge * hpge * hpge * cubic_hpge[i]*hpge_bin*hpge_bin*hpge_bin + 
					hpge * hpge * quadratic_hpge[i]*hpge_bin*hpge_bin + 
					hpge * slopes_hpge[i]*hpge_bin + 
					offsets_hpge[i];

				/* HPGe Baseline correct */
				hpge_e[i] = hpge_e[i]-hpge_b[i];
			}

		}
	}
#endif

	void calibrate() {
		for(int i = 0; i < 4; i++) {
			double labr_h = (double)(labr_h_e[i]) + r.Rndm() - 0.5;
			double labr_l = (double)(labr_l_e[i]) + r.Rndm() - 0.5;
			double labr_sis = (double)(labr_sis_e[i] - labr_sis_b[i]) + r.Rndm() - 0.5;
			double hpge   = (double)(hpge_e[i] - hpge_b[i]) + r.Rndm() - 0.5;
			double hpge_bin = (double)(h_bins[8+i])/(double)(h_max[8+i]);

			labr_h_e_calib[i] = (labr_h * labr_h * labr_h * cubic_labr_h[i]) + 
				(labr_h * labr_h * quadratic_labr_h[i]) + 
				(labr_h * slopes_labr_h[i]) + 
				offsets_labr_h[i];
			labr_sis_e_calib[i] = (labr_sis * labr_sis * labr_sis * cubic_labr_h[i]) + 
				(labr_sis * labr_sis * quadratic_labr_h[i]) + 
				(labr_sis * slopes_labr_h[i]) + 
				offsets_labr_h[i];

			labr_l_e_calib[i] = labr_l * slopes_labr_l[i] + offsets_labr_l[i];

			hpge_e_calib[i] = hpge * hpge * hpge * cubic_hpge[i]*hpge_bin*hpge_bin*hpge_bin + 
				hpge * hpge * quadratic_hpge[i]*hpge_bin*hpge_bin + 
				hpge * slopes_hpge[i]*hpge_bin + 
				offsets_hpge[i];

			hpge_e[i] = hpge_e[i] - hpge_b[i];
		}
	}

	void refreshTree() {
		int tries = 4;
		int keys_readable = 0;
		struct flock fl;
		int fd = f->GetFd();
		int ret;

		fl.l_whence = SEEK_SET;
		fl.l_start = 0;
		fl.l_len = 1;

		/*printf("Refreshing tree...\n");*/

		while(tries-- && (keys_readable == 0)) {
			/*flock(fd, LOCK_EX);*/
			fl.l_type = F_RDLCK;
			ret = fcntl(fd, F_SETLKW, &fl);
			if (ret == -1) {
				if (errno == EINTR) {
					continue;
				} else {
					perror("fcntl(F_SETLKW, F_RDLCK");
				}
			}

			/*printf("Got file lock.\n");*/

			if(ntuple->fDirectory->ReadKeys() != 0) keys_readable = 1;

			/*flock(fd, LOCK_UN);*/
			/*fl.l_type = F_UNLCK;
			  ret = fcntl(fd, F_SETLK, &fl);
			  if (ret == -1) {
			  perror("fcntl(F_SETLK, F_UNLCK");
			  }*/

			usleep(100);
		}
		if(tries == 0) {
			printf("Tried 10 times. Resetting...");
			// Reset, if we cannot find our keys
			f->Close();
			delete f;
			f = 0;
			ntuple = 0;
			first = 0;
			exit(1);
		}

		/*flock(fd, LOCK_EX);*/
		/*fl.l_type = F_RDLCK;
		  ret = fcntl(fd, F_SETLKW, &fl);
		  if (ret == -1) {
		  if (errno == EINTR) {
		  continue;
		  } else {
		  perror("fcntl(F_SETLKW, F_RDLCK");
		  }
		  }*/

		/*printf("Refreshing");*/
		ntuple->Refresh();

		/*flock(fd, LOCK_UN);*/
		fl.l_type = F_UNLCK;
		ret = fcntl(fd, F_SETLK, &fl);
		if (ret == -1) {
			perror("fcntl(F_SETLK, F_UNLCK");
		}

		if(first == ntuple->GetEntries()) {
			return;
		} 
	}

	void extendTree(int showProgress = 0) {
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;

		char friendname[255];
		filename[strlen(filename)-5] = '\0';
		sprintf(friendname,"%s_calib.root",filename);
		printf("friendname = %s\n",friendname);

		TFile* cal = new TFile(friendname,"recreate");
		TTree* h102 = new TTree("h102","SYNC level tree");

#if 0
		for(int i = 0; i < 8; ++i) {
			char buf[255];
			sprintf(buf,"%s/F",bc_names[i]);
			h102->Branch(bc_names[i],&labr_h_e_calib[i],buf);
			if(i < 4) {
				sprintf(buf,"%s/F",bc_names[8+i]);
				h102->Branch(bc_names[8+i],&hpge_e_calib[i],buf);  
			}
		}
#endif
		for(int i = 0; i < 4; ++i) {
			char buf[255];
			sprintf(buf,"%s/F",bc_names[i]);
			h102->Branch(bc_names[i],&labr_sis_e_calib[i],buf);
			if(i < 4) {
				sprintf(buf,"%s/F",bc_names[8+i]);
				h102->Branch(bc_names[8+i],&hpge_e_calib[i],buf);  
			}
		}

#ifdef USE_STEPWISE_CALIB
		int labr_last_calibration = 
		    (sizeof(labr_timestamp_start)/sizeof(labr_timestamp_start[0]))-1;
		int labr_current_calibration = 0;
#endif 

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		ntuple->SetBranchStatus("TIMESTAMP",1);
		for(int i = 0; i < 8; ++i) {  
			ntuple->SetBranchStatus(b_names[i*2],1);
			//ntuple->SetBranchStatus(b_names[i*2+1],1);
			if(i < 5) {
				ntuple->SetBranchStatus(b_names[64+i],1);
			}
			if(i < 4) {
				ntuple->SetBranchStatus(b_names[16+i*2],1);
				ntuple->SetBranchStatus(b_names[16+i*2+1],1);
			}
		}

		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {
			ntuple->GetEntry(entry);
			uint64_t timestamp;

			/* Build 64 bit timestamp */
			timestamp = ((uint64_t)time_hi) << 32 | time_lo;
			
			if(entry == first) ts_first = timestamp;

			if(showProgress == 1 && ++status_update_counter == 100000) {
				double progress = ((double)(entry-first))/((double)nof_entries);
				fprintf(stderr,"Filling histograms: "GREEN"%8d"BLACK"("GREEN
				    "%.1f%%"BLACK")\r",entry+1,progress*100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

#ifdef USE_STEPWISE_CALIB
			// Set current calibration
			if(labr_current_calibration < labr_last_calibration 
			    && timestamp > labr_timestamp_start[labr_current_calibration+1]) {
				++labr_current_calibration;
				printf("\nChanging labr calibration to %d starting at timestamp %d\n",
				    labr_current_calibration,timestamp);
			}
			calibrate_stepwise(labr_current_calibration);
#else
			calibrate();
#endif

			h102->Fill();

		}

		h102->Write();
		cal->Close();
	}

	typedef struct recalibration_set_s {
		int det;
		int from;
		int to;
		int t_from;
		int t_to;
		int n_peaks;
		float offset;
		float slope;
	} recalibration_set;

	void recalibrateTree(int showProgress = 0, int n_peaks = 2) {
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;

		char friendname[255];
		char recalibration_filename[255];

		filename[strlen(filename)-5] = '\0';
		sprintf(friendname,"%s_recalib.root",filename);
		printf("friendname = %s\n",friendname);

		sprintf(recalibration_filename,"%s_recalib.dat",filename);
		printf("recalibration = %s\n",recalibration_filename);
		printf("commencing recalibration with %d peaks.\n",n_peaks);

		TFile* cal = new TFile(friendname,"recreate");
		TTree* tree = new TTree("h103","SYNC level tree (recalibrated HPGE)");

		// Read recalibration file
#define MAX_CHARS 255
		FILE* recalibration_file = 0;
		char line[MAX_CHARS];

		recalibration_file = fopen(recalibration_filename,"r");
		if(recalibration_file == 0) {
			printf("Error opening recalibration file: %s\n",recalibration_filename);
			exit(1);
		}

		// Pre-allocate 100 records and 4 detectors
		recalibration_set recal_data[4][100];
		size_t n_records[4];
		for (int i = 0; i < 4; ++i) { n_records[i] = 0;}  

		// Go through the lines
		while(fgets(line,sizeof(line),recalibration_file) != 0) {
			int det = 0;
			int from = 0, to = 0;
			int t_from = 0, t_to = 0;
			float* heights;
			float* energies;
			float* widths;
			float offset = 0;
			float slope = 0;
			//float dummy;
			char* str = 0;

			// Allocate memory for the data 
			heights = (float*) malloc(n_peaks * sizeof(float));
			energies = (float*) malloc(n_peaks * sizeof(float));
			widths = (float*) malloc(n_peaks * sizeof(float));

			// Parse line
			str = strtok(line," ");   if(str != 0) sscanf(str,"%d",&det);
			str = strtok(0," ");   if(str != 0) sscanf(str,"%d",&from);
			str = strtok(0," ");   if(str != 0) sscanf(str,"%d",&to);
			str = strtok(0," ");   if(str != 0) sscanf(str,"%d",&t_from);
			str = strtok(0," ");   if(str != 0) sscanf(str,"%d",&t_to);

			for(int i = 0; i < n_peaks; ++i) {
				str = strtok(0," ");   if(str != 0) sscanf(str,"%f",&heights[i]);
				str = strtok(0," ");   if(str != 0) sscanf(str,"%f",&energies[i]);
				str = strtok(0," ");   if(str != 0) sscanf(str,"%f",&widths[i]);
			}

			str = strtok(0," ");   if(str != 0) sscanf(str,"%f",&offset);
			str = strtok(0," ");   if(str != 0) sscanf(str,"%f",&slope);

			// Test output
			printf("%lu %d %d %d %f %f\n", n_records[det-1], det, from, to, offset, 
			    slope);
			for(int i = 0; i < n_peaks; ++i) {
				printf("  %f %f %f\n", energies[i], heights[i], widths[i]);
			}fflush(stdout);

			printf("a\n");fflush(stdout);
			// Store data
			recal_data[det-1][n_records[det-1]].det = det;
			recal_data[det-1][n_records[det-1]].from = from;
			recal_data[det-1][n_records[det-1]].to = to;
			recal_data[det-1][n_records[det-1]].t_from = t_from;
			recal_data[det-1][n_records[det-1]].t_to = t_to;
			recal_data[det-1][n_records[det-1]].offset = offset;
			recal_data[det-1][n_records[det-1]].slope = slope;

			printf("b\n");fflush(stdout);

			free(heights);
			free(widths);
			free(energies);
			++n_records[det-1];

			printf("c\n");fflush(stdout);
		}
		fclose(recalibration_file);

		// Create new branches 
		for(int i = 0; i < 4; ++i) {
			char buf[255];
			sprintf(buf,"%s/F",bcr_names[i]);
			tree->Branch(bcr_names[i],&labr_sis_e_calib[i],buf);
			if(i < 4) {
				sprintf(buf,"%s/F",bcr_names[8+i]);
				tree->Branch(bcr_names[8+i],&hpge_e_calib[i],buf);  
			}
		}

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		ntuple->SetBranchStatus("TIMESTAMP",1);
		for(int i = 0; i < 8; ++i) {  
			ntuple->SetBranchStatus(b_names[i*2],1);
			if(i < 5) {
				ntuple->SetBranchStatus(b_names[64+i],1);
			}
			if(i < 4) {
				ntuple->SetBranchStatus(b_names[16+i*2],1);
				ntuple->SetBranchStatus(b_names[16+i*2+1],1);
			}
		}

		// Loop over tree 
		int cal_idx = 0;
		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {
			uint64_t timestamp;

			// Update recal data
			while(entry >= recal_data[0][cal_idx].to && cal_idx < (int)(n_records[0])-1) {
				++cal_idx;
				printf("Using calibration data from %d to %d "
				    "(offset=%.2f, slope=%.5f)\n",
				    recal_data[0][cal_idx].from, recal_data[0][cal_idx].to,
				    recal_data[0][cal_idx].offset, recal_data[0][cal_idx].slope);
			}

			ntuple->GetEntry(entry);
			
			/* Build 64 bit timestamp */
			timestamp = ((uint64_t)time_hi) << 32 | time_lo;
			if(entry == first) ts_first = timestamp;

			// Status update
			if(showProgress == 1 && ++status_update_counter == 100000) {
				double progress = ((double)(entry-first))/((double)nof_entries);
				fprintf(stderr,"Filling histograms: "GREEN"%8d"BLACK"("GREEN
					"%.1f%%"BLACK")\r",entry+1,progress*100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

			calibrate(); // Do the normal calibration

			// Recalibrate
			for(int i = 0; i < 4; ++i) {
				hpge_e_calib[i] = (hpge_e_calib[i] * recal_data[i][cal_idx].slope) 
				    + recal_data[i][cal_idx].offset;
			}


			tree->Fill();

		}

		tree->Write();
		cal->Close();
	}

	void showScaler(int showProgress = 0, int showScalers = 0) {
		// Manual loop
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;
		char run_nr[8];
		char buf[256];
		char path[256];
		char* path_ptr = &(path[0]);
		char* filename_ptr = &(filename[0]);

		Int_t entries = ntuple->GetEntries();

		/* Open output file */
		FILE* sc_out = 0;

		strcpy(path,filename);
		path_ptr = dirname(path);
		filename_ptr = basename(filename);
		PRINTD("%s, path_ptr: %s, file_ptr: %s\n",filename,path_ptr,filename_ptr); fflush(stdout);
		PRINTD("%s, path: %s, file: %s\n",filename,path,filename); fflush(stdout);

		sprintf(buf,"%s",path_ptr);
		gSystem->MakeDirectory(buf);

		if(strcmp(filename_ptr,"stream.root") == 0) {
			sprintf(buf,"%s/stream_scalers.dat",path_ptr);
		} else {
			if(strncmp(filename_ptr,"run",3) == 0) {
				strncpy(run_nr,(filename_ptr + 3),3); run_nr[3] = '\0';
				sprintf(buf,"%s/run%s_scalers.dat",path,run_nr);
			} else {
				printf("No stream file, and no run number given. Saving tv spectra as unknown.\n");
				sprintf(buf,"%s/unknown_scalers.dat",path);
			}
		}
		PRINTD("Opening scaler output file: %s\n",buf);
		sc_out = fopen(buf,"w");
		if(!sc_out) {
			printf("Could not open scaler output file '%s'\n.",buf);
		}

		// Activate only necessary branches
		ntuple->SetBranchStatus("*",0);
		/*ntuple->SetBranchStatus("TIMESTAMP",1);*/
		for(int i = 0; i < 8; ++i) {  
			ntuple->SetBranchStatus(b_names[44+i],1);
			ntuple->SetBranchStatus(b_names[72+i],1);
			if(i < 4) {
				ntuple->SetBranchStatus(b_names[52+i],1);
				ntuple->SetBranchStatus(b_names[80+i],1);
			}
		}
		for(int i = 0; i < 64; ++i) {
			ntuple->SetBranchStatus(trlo_names[i],1);
		}

		/* Loop through scaler events */ 
		fprintf(sc_out,"# scaler1_raw scaler1_before_deadtime scaler1_after_deadtime scaler1_after_reduction ...\n");
    int fi = 1;
		for(int entry = first; entry < entries; ++entry) {
			if(showProgress == 1 && ++status_update_counter == 100000) {
				double progress = ((double)(entry-first))/((double)nof_entries);
				fprintf(stderr,"Computing statistics: "GREEN"%8d"BLACK"("GREEN
				    "%.1f%%"BLACK")\r",entry+1,progress*100);
				fflush(stdout);
				status_update_counter = 0;
			}
			if(entry == entries-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

			b_trlo_sc[0]->GetEntry(entry);

			if(trlo_sc[0].raw == 0) continue;
			ntuple->GetEntry(entry);


      if (fi == 1) {
        for(int i = 0; i < 16; ++i) {
          trlo_sc_start[i].raw = trlo_sc[i].raw;
          trlo_sc_start[i].bdt = trlo_sc[i].bdt;
          trlo_sc_start[i].adt = trlo_sc[i].adt;
          trlo_sc_start[i].ard = trlo_sc[i].ard;
        }
        fi = 0;
      }

			/* Increase scalers */
			for(int i = 0; i < 16; ++i) {
				trlo_sc_sum[i].raw = trlo_sc[i].raw - trlo_sc_start[i].raw;
				trlo_sc_sum[i].bdt = trlo_sc[i].bdt - trlo_sc_start[i].bdt;
				trlo_sc_sum[i].adt = trlo_sc[i].adt - trlo_sc_start[i].adt;
				trlo_sc_sum[i].ard = trlo_sc[i].ard - trlo_sc_start[i].ard;
			}

			// Scalers update
#define MAX_SCALER_26BIT 0x4000000
			if(labr_s[0] != 0) {
				for(int i = 0; i < 8; ++i) {
					if(labr_s_prev[i] != 0) {
						if (labr_s[i] > labr_s_prev[i]) {
							labr_rate[i] = labr_s[i]-labr_s_prev[i];
						} else {
							labr_rate[i] = (labr_s[i]+MAX_SCALER_26BIT)-(labr_s_prev[i]);
						}
					}
					if(labrl_s_prev[i] != 0) {
						if (labrl_s[i] > labrl_s_prev[i]) {
							labrl_rate[i] = labrl_s[i]-labrl_s_prev[i];
						} else {
							labrl_rate[i] = (labrl_s[i]+MAX_SCALER_26BIT)-(labrl_s_prev[i]);
						}
					}
					labr_s_prev[i] = labr_s[i];
					labrl_s_prev[i] = labrl_s[i];
					if(i < 4) {
						if(hpge_s_prev[i] != 0) {
							if (hpge_s[i] > hpge_s_prev[i]) {
								hpge_rate[i] = hpge_s[i]-hpge_s_prev[i];
							} else {
								hpge_rate[i] = (hpge_s[i]+MAX_SCALER_26BIT)-(hpge_s_prev[i]);
							}
						}
						if(hpgel_s_prev[i] != 0) {
							if (hpgel_s[i] > hpgel_s_prev[i]) {
								hpgel_rate[i] = hpgel_s[i]-hpgel_s_prev[i];
							} else {
								hpgel_rate[i] = (hpgel_s[i]+MAX_SCALER_26BIT)-(hpgel_s_prev[i]);
							}
						}
						if(hpge_crm_prev[i] != 0) {
							if (hpge_crm[i] > hpge_crm_prev[i]) {
								hpge_crm_rate[i] = hpge_crm[i]-hpge_crm_prev[i];
							} else {
								hpge_crm_rate[i] = (hpge_crm[i]+MAX_SCALER_26BIT)-(hpge_crm_prev[i]);
							}
						}
						hpge_s_prev[i] = hpge_s[i];
						hpgel_s_prev[i] = hpgel_s[i];
						hpge_crm_prev[i] = hpge_crm[i];
					}
				}
			}  

			/* Print scaler event */
			if(showScalers == 1) {
				printf("Trlo Scalers ..........................................."
				       "..........................\n");
				printf("Name               RAW #                      B.DT      "
				       "A.DT      A.RD Deadtime   \n");
				for(int i = 0; i < 16; ++i) {
					printf("%2d: %s %9d # %2d:%s %9d %9d %9d %.2f\n",i+1,
					    trlo_raw_names[i], trlo_sc[i].raw, i+1,trlo_tpat_names[i],
					    trlo_sc[i].bdt, trlo_sc[i].adt, trlo_sc[i].ard,
					    100*(double)(trlo_sc[i].adt)/(double)(trlo_sc[i].bdt));
				}
				printf("\n");
			}

			/* Print scaler event to file */
			for(int i = 0; i < 16; ++i) {
				fprintf(sc_out,"%d %d %d %d ",
						trlo_sc[i].raw, trlo_sc[i].bdt,
						trlo_sc[i].adt, trlo_sc[i].ard);
			}
			for(int i = 0; i < 8; ++i) {
				fprintf(sc_out,"%d %d ",labrl_rate[i], labr_rate[i]);
			}
			for(int i = 0; i < 4; ++i) {
				fprintf(sc_out,"%d %d %d ",hpgel_rate[i], hpge_rate[i],
				    hpge_crm_rate[i]);
			}
			fprintf(sc_out,"\n");

		}

		/* Print summary */
		printRealTime();
		getRunNumber(run_nr);
		printf("Trlo Scalers for run %s .............................................."
		       ".....................\n", run_nr);
		printf("Name                     RAW #                          B.DT          "
		       "A.DT          A.RD Deadtime  Reduction \n");
		for(int i = 0; i < 16; ++i) {
			printf("%2d: %s %15lld # %2d:%s %13lld %13lld %13lld    %5.2f %5.2f\n",i+1,
			    trlo_raw_names[i], (long long)trlo_sc_sum[i].raw, i+1, trlo_tpat_names[i],
			    (long long) trlo_sc_sum[i].bdt, (long long)trlo_sc_sum[i].adt,
			    (long long)trlo_sc_sum[i].ard, 
			    100*(1.-(double)(trlo_sc_sum[i].adt)/(double)(trlo_sc_sum[i].bdt)),
			    ((double)(trlo_sc_sum[i].adt)/(double)(trlo_sc_sum[i].ard)));
		}


	}

	void fillHistograms(int showProgress = 0, int only_tv = 0, int gui = 0) {
		// Manual loop
		int nof_entries = ntuple->GetEntries() - first;
		int status_update_counter = 0;
		int statusbar_update_counter = 0;
		uint64_t timestamp = 0;

		if(only_tv == 1) {
			// Activate only necessary branches
			ntuple->SetBranchStatus("*",0);
			ntuple->SetBranchStatus("TS_LO",1);
			ntuple->SetBranchStatus("TS_HI",1);
			for(int i = 0; i < 4; ++i) {  
				ntuple->SetBranchStatus(b_names[144+i],1);
				ntuple->SetBranchStatus(b_names[148+i],1);
				ntuple->SetBranchStatus(b_names[152+i],1);
				ntuple->SetBranchStatus(b_names[156+i],1);
			}
			ntuple->SetBranchStatus("TPAT",1);
		} else {
			// Deactivate trlo scalers
			//for(int i = 0; i < 64; ++i) {
			//  ntuple->SetBranchStatus(trlo_names[i],0);
			//}
		}

#ifdef USE_STEPWISE_CALIB
		int labr_last_calibration = (sizeof(labr_timestamp_start)/sizeof(labr_timestamp_start[0]))-1;
		int labr_current_calibration = 0;
		//double labr_last_timestamp = labr_timestamp_start[labr_current_calibration];
#endif

		int cnt = 0;
		printf("Filling %d entries.\n", nof_entries);
		for(int entry = first; entry < ntuple->GetEntries(); ++entry) {
			ntuple->GetEntry(entry);

			/* Build 64 bit timestamp */
			timestamp = ((uint64_t)time_hi) << 32 | time_lo;

			if(entry == first) ts_first = timestamp;


			if(showProgress == 0 && first == 0 && (++status_update_counter == 100000)) {
				status_update_counter = 0;
				double progress = (double)(entry-first)/(double)(nof_entries);
				printf("\rCatching up: "GREEN"%.1f%%"BLACK"                   ",
						100*progress); fflush(stdout);
			}

			/* Allow to have interaction while filling histograms */
			cnt++;
			if(first != 0 && cnt == 100)
			{
				gSystem->ProcessEvents();
				cnt = 0;
			}
			if(first == 0 && cnt == 10000)
			{
				gSystem->ProcessEvents();
				cnt = 0;
			}

			if(first == 0 && entry == first) {
				timestamp_begin = timestamp;
			}


#ifdef USE_STEPWISE_CALIB
			// Write tv spectra and clear histograms
			if(labr_current_calibration < labr_last_calibration 
					&& timestamp > labr_timestamp_start[labr_current_calibration+1]) {
				++labr_current_calibration;
				printf("\nChanging labr calibration to %d starting at timestamp %d\n",
						labr_current_calibration,timestamp);
				writeTv(labr_current_calibration+1);
				clearHistograms();
			}
#endif

			if(gui == 1 && ++statusbar_update_counter == 10000) {
				char buf[255];
				statusbar_update_counter = 0;
				if(first == 0) {
					double progress = (double)(entry-first)/(double)(nof_entries);
					sprintf(buf,"Catching up: %.1f      ",100*progress);
				} else {
					sprintf(buf,"Events: %lld",ntuple->GetEntries());
				}
				status->SetText(buf,0);
				gPad->Modified();
				gPad->Update();
				gSystem->ProcessEvents();
			}

			if(showProgress == 1 && ++status_update_counter == 100000) {
				double progress = ((double)(entry-first))/((double)nof_entries);
				fprintf(stderr,
						"Filling histograms: "GREEN"%8d"BLACK"("GREEN"%.1f%%"BLACK")\r",
						entry+1,progress*100);
				fflush(stderr);
				status_update_counter = 0;
			}
			if(entry == ntuple->GetEntries()-1) { fprintf(stderr,"\n"); }

			PRINTD("%d\n",entry);

			if(checkCuts() > 0) continue;

#ifdef USE_STEPWISE_CALIB
			calibrate_stepwise(labr_current_calibration);
#else
			calibrate();
#endif

			int ref = 0;
			int mul_g = 0, mul_h = 0;
			float t_rf = trig_t[11]/8.;
			float t_ref = labrl_t[ref]/8.;

			// Labr spectra
			for(int i = 0; i < 4; ++i) {
				int labr_e = labr_sis_e[i] - labr_sis_b[i];
				if((USED_CONDITION) && labr_e > thr_labr_sis[i]) {
					/* LaBr energies */
					h[i]->Fill(labr_e);
					h[i+4]->Fill(labr_l_e[i]);

					float time_l = labrl_t[i]/8.;
					float time_h = labr_t[i]/8.;
					if(only_tv == 0) {
						/* Calibrated LaBr energies */
						if(((tpat & TPAT_LABR_SINGLE_H) > 0 
						    || (tpat & (1<<2)) > 0 
						    || (tpat & (1<<3)) > 0 
						    || (tpat & (1<<12)) > 0
						    || (tpat & (1<<13)) > 0) 
						    && labr_t[i] > 0) {
							hc[i]->Fill(labr_sis_e_calib[i]);
							if(i < 4) {
								float tdiff = time_l - t_rf;

								// Condition 1
								if(time_l > 0 && t_rf > 0 
								    && tdiff > labr_w1[2*i]
								    && tdiff < labr_w1[(2*i)+1])
								    //&& tdiff > labr_t_min[i] 
								    //&& tdiff < labr_t_max[i])
								{
									hc[16+i]->Fill(labr_sis_e_calib[i]);
								}
								// Condition 2
								if(time_l > 0 && t_rf > 0 
								    && tdiff > labr_w2[(2*i)]
								    && tdiff < labr_w2[(2*i)+1])
								    //&& tdiff > labr_t_min[i] 
								    //&& tdiff < labr_t_max[i])
								{
									hc[24+i]->Fill(labr_sis_e_calib[i]);
								}
							}
						}

						// Raw time spectra
						if(time_l > 0) {
							int all_others = 0;
							float t_trig = trig_t[0]/8.;
							float tdiff = time_h - t_trig;

							(void)tdiff;

							for(int j = 0; j < 4; ++j)
								if(j!=i && labrl_t[j] > 0)
									all_others = 1;
							for(int j = 0; j < 4; ++j) 
								if(hpgel_t[j] > 0) 
									all_others = 1;

							if (0) { printf("%d",all_others); }

							//if(all_others == 0)
							//	h[48+i]->Fill(time_l); /* RAW LOW */
							//if(all_others == 0) 
							//	h[48+i]->Fill(time_h); /* RAW HIGH */
							//if(trig_t[0] > 0 && all_others == 0)
							//	h[48+i]->Fill(tdiff); /*Self-trigger*/

							if(t_ref > 0) {
								all_others = 0;
								for(int j = 0; j < 4; ++j) 
									if(j != ref && j != i 
											&& labrl_t[j] > 0)
										all_others = 1;
								for(int j = 0; j < 4; ++j)
									if(hpgel_t[j] > 0) 
										all_others = 1;

								/* vs. LaBr1 LOW */
								if(all_others == 0)
									h[48+i]->Fill(time_l - t_ref);
								/* vs. LaBr1 HIGH */
								//if(all_others == 0) 
								//	h[48+i]->Fill(time_h - t_ref);
							}
						}

						// Individual time spectra
						//if(labr_t[i]>0 && labr_t[0]>0) { 
						//	/* LABR vs. Labr 1 HIGH */
						//	h[12+i]->Fill(time_l - t_ref);
						//}
						if(time_h > 0 && t_rf > 0 
						    && labr_sis_e_calib[i]>1000. /* LABR vs RF */
						    //&& (time_h - t_rf) > labr_t_min[i] 
						    //&& (time_h - r_rf) < labr_t_max[i])
							) {
								h[12+i]->Fill(time_h - t_rf);
							}
						h[38]->Fill(i+1);
						++mul_g;

						// Correlation
						for(int j = i; j < 8; ++j) {
							if(labr_sis_e[j] > thr_labr_sis[j]) {
								h2[1]->Fill(i+1,j+1);
							}
						}
						for(int j = 0; j < 4; ++j) {
							if(hpge_e[j] > thr_hpge[j]) {
								h2[1]->Fill(i+1,j+8+1);
							}
						}
					}
				}
			}
			// HPGe spectra
			for(int i = 0; i < 4; ++i) {
			
				/* Energies */
				if(hpge_e[i] > thr_hpge[i]) {
					h[8+i]->Fill(hpge_e[i]);
					if(only_tv == 0) {
						float time_h = hpge_t[i]/8.;
						float time_l = hpgel_t[i]/8.;
						float t_trig = trig_t[2]/8.;
						float t_diff = time_h - t_trig;
						float t_diff_l = time_l - t_trig;
						float t_diff_ref = time_l - t_ref;
						float t_diff_rf = time_l - t_rf;

						(void)t_diff;
						(void)t_diff_l;

						// Cut on HPGE OR trigger, a valid time 
						// in this channel, and no pileup
						if((tpat & TPAT_HPGE_SINGLE_H) > 0 
						     && time_l > 0 /*&& hpge_pileup[i] == 0*/
								/*&& t_diff > hpge_t_min[i] 
								  && t_diff < hpge_t_max[i])*/
							) {
								hc[8+i]->Fill(hpge_e_calib[i]);
								h2[8+i]->Fill(timestamp
								    -timestamp_begin,hpge_e_calib[i]);
							}
						if(time_l > 0) {
							// Raw time spectra
							int all_others = 0;
							for(int j = 0; j < 4; ++j) 
								if(j != i && hpgel_t[j] > 0)
									all_others = 1;
							for(int j = 0; j < 4; ++j)
								if(labrl_t[j] > 0)
									all_others = 1;

							if(0) printf("%d",all_others);

							/* Raw low time */
							//h[48+8+i]->Fill(time_l);
							//h[48+8+i]->Fill((trig_t[i]-trig_t[0])/8.);
							/* Raw high time */
							//h[h_htr+i]->Fill(time);

							/* Self-trigger LOW */
							//if(t_trig > 0)
							//	h[48+8+i]->Fill(t_diff_l);

							/* vs. LaBr 1 LOW */
							//if(t_ref > 0)
							//	h[48+8+i]->Fill(t_diff_ref);
							
							/* vs. RF time */
							if(t_rf > 0)
								h[48+8+i]->Fill(t_diff_rf);
							
							// Individual time spectra (2D)
							if(t_rf > 0){
								h2[4+i]->Fill(t_diff_rf, 
										hpge_e_calib[i]);
							}
						}
						/*if(hpge_t[i] > 0 && labr_t[0] > 0 ) */
						if(hpge_t[i] > 0 && trig_t[3] > 0 
								&& hpge_e_calib[i] > 1000.) {
							//h[20+i]->Fill(t_diff_ref);
							h[20+i]->Fill(t_diff_rf);
						}
						h[38]->Fill(8+i+1);
						++mul_h;

						// Correlation
						for(int j = i; j < 4; ++j) {
							if(hpge_e_calib[j] > thr_hpge[j]) {
								h2[1]->Fill(i+8+1,j+8+1);
							}
						}
					}
				}
			}
			if(only_tv == 0) {
				// LaBr sum
				for(int i = 0; i < 8; ++i) {
					if(labr_sis_e[i] > thr_labr_sis[i]
					    && labr_sis_e_calib[i] > 60) {
						h[32+0]->Fill(labr_sis_e_calib[i]);
					}
				}
				// HPGe sum
				for(int i = 0; i < 4; ++i) {
					if((hpge_e[i] > thr_hpge[i]) 
							&& hpgel_t[i] > 0 
							/*&& hpge_pileup[i] == 0*/)
					{
						h[32+1]->Fill(hpge_e_calib[i]);
					}
				}
				// Special Spectra

				// Coincidence Time
				//if(trig_t[1] > 0 && trig_t[0] > 0) {
				//  h[32+2]->Fill((trig_t[1] - trig_t[0])*0.125);
				//}

				// Coincidence matrix HPGe LaBr
				for(int i = 0; i < 4; ++i ) {
					float t_labr = labrl_t[i]/8.;
					float t_labr_rf = t_labr - t_rf;
					if((tpat & TPAT_COINCIDENCE_LL)) {
						for(int j = 0; j < 4; ++j) {
							float t_hpge = hpgel_t[j]/8.;
							float t_hpge_rf = t_hpge - t_rf;

							(void)t_hpge_rf;
							//if(/*hpge_e[j] < thr_hpge[j]*/) continue;
							/* Use coincidence windows */
							if(t_labr_rf > labr_w1[2*i] &&
							    (t_labr_rf < labr_w1[(2*i)+1])
							    && (t_hpge_rf > hpge_w1[2*j]
							    && (t_hpge_rf < hpge_w1[(2*j)+1])))
							    //|| (labr_h_e_calib[i] < 150)
							    /*&& hpge_e_calib[j] > 1000)*/
							{
								h2[0]->Fill(labr_sis_e_calib[i],
								    hpge_e_calib[j]);
							}
						}
					}
					// Labr-Labr
					if((tpat & TPAT_LABR_M)) {
						for(int j = i+1; j < 4; ++j) {
							float t_labr2 = labr_t[j]/8.;
							float t_labr2_rf = t_labr2 - t_rf;
							//if(labr_h_e[j] < thr_labr_h[j] /*
							//    || labr_h_e_calib[i] < 550*/) continue;
							if ((t_labr_rf > labr_w1[2*i])
							    && (t_labr_rf < labr_w1[(2*i)+1])  
							    && (t_labr2_rf > labr_w1[2*j])
							    && (t_labr2_rf < labr_w1[(2*j)+1]))
							    //||
							    //(labr_h_e_calib[i] < 150)
							    //&& labr_h_e_calib[j] > 1500)
							
							{
								h2[2]->Fill(labr_sis_e_calib[i],
								    labr_sis_e_calib[j]);
								h2[2]->Fill(labr_sis_e_calib[j],
								    labr_sis_e_calib[i]);
							}
						}
					}
				}
			}

			if(only_tv == 0) {
				// Multiplicity histograms
				h[35+0]->Fill(mul_g);
				h[35+1]->Fill(mul_h);
				h[35+2]->Fill(mul_g+mul_h);

				/* Update Scaler current values */
				if(labr_s[0] != 0 || labrl_s[0] != 0) {
					for(int i = 0; i < 8; ++i) {
						labr_s_cur[i] = labr_s[i];
						labrl_s_cur[i] = labrl_s[i];
						if(i < 4) {
							hpge_s_cur[i] = hpge_s[i];
							hpgel_s_cur[i] = hpgel_s[i];
							hpge_crm_cur[i] = hpge_crm[i];
						}
					}
				}  
			}
		}
		
		/* Update scaler rates */
		if (only_tv == 0) {
			for (int i = 0; i < 8; ++i) {
				h[39+0]->SetBinContent(i+1, labr_s_cur[i] - labr_s_prev[i]);
	
				labr_rate[i] = labr_s_cur[i] - labr_s_prev[i];
				labrl_rate[i] = labrl_s_cur[i] - labrl_s_prev[i];
		
				labr_s_prev[i] = labr_s_cur[i];
				labrl_s_prev[i] = labrl_s_cur[i];
			}
			for (int i = 0; i < 4; ++i) {
				h[39+1]->SetBinContent(i+1, hpge_s_cur[i] - hpge_s_prev[i]);
			
				hpge_rate[i] = hpge_s_cur[i] - hpge_s_prev[i];
				hpgel_rate[i] = hpgel_s_cur[i] - hpgel_s_prev[i];
				hpge_crm_rate[i] = hpge_crm_cur[i] - hpge_crm_prev[i];
				
				hpge_s_prev[i] = hpge_s_cur[i];
				hpgel_s_prev[i] = hpgel_s_cur[i];
				hpge_crm_prev[i] = hpge_crm_cur[i];
			}
		}

	}

	void printRealTime(int gui = 0) {
		float realtime;

		ntuple->SetBranchStatus("TS_HI",1);
		ntuple->SetBranchStatus("TS_LO",1);

		// Print real time
		ntuple->GetEntry(0);
		start = (((uint64_t)time_hi) << 32) | time_lo;
		ntuple->GetEntry(ntuple->GetEntries()-1);
		stop = (((uint64_t)time_hi) << 32) | time_lo;

		realtime = (stop - start) / 1e8;

		if(gui == 1) {
			char buf[255];
			sprintf(buf,"Real time: %.2f s", realtime);
			status->SetText(buf,1);
		}
    printf("Start: %lu\n", start);
    printf("Stop: %lu\n", stop);
		printf("Real time: %.2f s.\n", realtime);
	}

	void draw() {
		int showProgress = 0;
		int only_tv = 0;
		int gui = 1;
		uint64_t timestamp = 0;

		refreshTree();

		// Clear rates histogram
		h[38]->Reset();

		fillHistograms(showProgress, only_tv, gui);

		printRealTime(gui);

		// Print scalers
		printf(  "--- Scalers ----------- 1 sec ---------\n");
		printf(  " Name   | Rate Lo Thr | Rate Hi Thr    \n");
		for(int i = 0; i < 8; ++i) {
			printf(" LaBr %d | %6d       %6d               \n",
			    i+1, labrl_rate[i], labr_rate[i]);
		}
		for(int i = 0; i < 4; ++i) {
			printf(" HPGe %d | %6d       %6d   CRM: %6d    \n",
			    i+1, hpgel_rate[i], hpge_rate[i], hpge_crm_rate[i]);
		}

		// Scale rate histogram
		ts_last = timestamp;
		int tdiff = ts_last - ts_first;
		if(tdiff <= 0) tdiff = 1;
		h[38]->Scale(1./tdiff);

		gSystem->ProcessEvents();

		first = (Int_t)ntuple->GetEntries();

		redrawHistograms();

		gSystem->ProcessEvents();

		writeTv();

	}

	void redrawHistograms() {
		TCanvas* c[10];
		TCanvas* c1 = ec_l->GetCanvas();
		TCanvas* c2 = ec_h->GetCanvas();
		TCanvas* c3 = ec_3->GetCanvas();
		TCanvas* c4 = ec_4->GetCanvas();
		TCanvas* c5 = ec_5->GetCanvas();
		TCanvas* c6 = ec_6->GetCanvas();
		TCanvas* c7 = ec_7->GetCanvas();
		TCanvas* c8 = ec_8->GetCanvas();
		TCanvas* c9 = ec_9->GetCanvas();
		c[9] = ec[9]->GetCanvas();

		// Redraw histograms
		for(int i=0; i < 4; ++i) {
			c1->cd(i+1);
			h[i]->Draw();
			c6->cd(i+1);
			hc[i]->Draw();
			c5->cd(i+1);
			h[12+i]->Draw();
			c8->cd(i+1);
			h[48+i]->Draw();
		}
		for(int i=0; i < 4; ++i) {
			c2->cd(i+1);
			h[8+i]->Draw();
			c7->cd(i+1);
			hc[8+i]->Draw();
			c2->cd(4+i+1);
			h2[4+i]->Draw("col");
			c[9]->cd(i+1);
			h2[8+i]->Draw("col");
			c8->cd(4+i+1);
			h[48+8+i]->Draw();
			c9->cd(i+1);
			hc[16+i]->Draw();
			c9->cd(4+i+1);
			hc[24+i]->Draw();
		}


		c3->cd(1); h[32+0]->Draw();
		c3->cd(2); h[32+1]->Draw();
		c3->cd(3); h2[2]->Draw("col");
		c3->cd(4); h2[0]->Draw("col");
		c4->cd(1); h[35+0]->Draw("");
		c4->cd(2); h[35+1]->Draw("");
		c4->cd(3); h[35+2]->Draw("");
		c4->cd(4); h[35+3]->Draw("");
		c4->cd(5); h[39+0]->Draw("");
		c4->cd(6); h[39+1]->Draw("");
		c4->cd(7); h2[1]->Draw("col");
		c4->cd(8); h[42+4]->Draw();

		// Update the canvas we are looking at
		switch(tab->GetCurrent()) {
			case 0:
				c1->Update();
				break;
			case 1:
				c5->Update();
				break;
			case 2:
				c6->Update();
				break;
			case 3:
				c9->Update();
				break;
			case 4:
				c2->Update();
				break;
			case 5:
				c7->Update();
				break;
			case 6:
				c3->Update();
				break;
			case 7:
				c4->Update();
				break;
			case 8:
				c8->Update();
				break;
			case 9:
				c[9]->Update();
		}

	}

	void getRunNumber(char* run_nr) {
		char* filename_ptr = &(filename[0]);
		filename_ptr = basename(filename);
		strncpy(run_nr,(filename_ptr + 3),3); run_nr[3] = '\0';
	}

	void fillCalibArrays() {
		for(int i = 0; i < 8; ++i) {
			calib_offsets[i] = offsets_labr_h[i];
			calib_slopes[i] = slopes_labr_h[i];
			calib_quad[i] = quadratic_labr_h[i];
			calib_cubic[i] = cubic_labr_h[i];
		}
		for(int i = 0; i < 4; ++i) {
			calib_offsets[i+8] = offsets_hpge[i];
			calib_slopes[i+8] = slopes_hpge[i];
		}
		for(int i = 0; i < 5; ++i) {
			calib_offsets[i+12] = offsets_mesytec[i];
			calib_slopes[i+12] = slopes_mesytec[i];
		}
	}


	void clearHistograms() {
		int h_map[] = {0,1,2,3,4,5,6,7,8,9,10,11,42,43,44,45,46}; // List of histograms
		for(int i = 0; i < (int)(sizeof(h_map)/sizeof(int)); ++i) {
			h[h_map[i]]->Reset();
		}
	}

	void writeTv(int index = 0) {
		// Write TV spectra to disk (with an appended 'index', if index != 0)
		FILE* tv_out;
		int h_map[] = {0,1,2,3,4,5,6,7,8,9,10,11,42,43,44,45,46}; // List of histograms
		int n_map[] = {0,1,2,3,4,5,6,7,8,9,10,11,36,37,38,39,40}; // List of names
		for(int i = 0; i < (int)(sizeof(h_map)/sizeof(int)); ++i) {
			char buf[256];
			char calbuf[256];
			char path[256];
			char run_nr[8];
			char* path_ptr = &(path[0]);
			char* filename_ptr = &(filename[0]);

			fillCalibArrays();

			strcpy(path,filename);
			path_ptr = dirname(path);
			filename_ptr = basename(filename);
			PRINTD("%s, path_ptr: %s, file_ptr: %s\n",
			    filename,path_ptr,filename_ptr); fflush(stdout);
			PRINTD("%s, path: %s, file: %s\n",filename,path,filename); fflush(stdout);

			sprintf(buf,"%s/tv",path_ptr);
			sprintf(calbuf,"%s/tv",path_ptr);
			gSystem->MakeDirectory(buf);

			if(strncmp(filename_ptr,"stream",6) == 0) {
				sprintf(buf,"%s/tv/%s_%s.tv",path_ptr,filename_ptr,h_names[n_map[i]]);
				sprintf(calbuf,"%s/tv/%s_%s.cal",path_ptr,filename_ptr,
				    h_names[n_map[i]]);
			} else {
				if(strncmp(filename_ptr,"run",3) == 0) {
					strncpy(run_nr,(filename_ptr + 3),3); run_nr[3] = '\0';
					if(index == 0) {
						sprintf(buf,"%s/tv/run%s_%s.tv",
						    path,run_nr,h_names[n_map[i]]);
						sprintf(calbuf,"%s/tv/run%s_%s.cal",
						    path,run_nr,h_names[n_map[i]]);
					} else {
						sprintf(buf,"%s/tv/run%s_%s_%d.tv",
						    path,run_nr,h_names[n_map[i]],index);
						sprintf(calbuf,"%s/tv/run%s_%s_%d.cal",
						    path,run_nr,h_names[n_map[i]],index);
					}
				} else {
					printf("No stream file, and no run number given. "
					    "Saving tv spectra as unknown.\n");
					sprintf(buf,"%s/tv/unknown_%s.tv",path,h_names[n_map[i]]);
					sprintf(calbuf,"%s/tv/unknown_%s.cal",path,h_names[n_map[i]]);
				}
			}
			PRINTD("Opening tv output file: %s\n",buf);
			tv_out = fopen(buf,"w");
			if(!tv_out) {
				printf("Could not open tv output file '%s'\n.",buf);
				continue;
			}

			fprintf(tv_out,"# Real time: %lu s.\n", stop - start);
			for(int n = 0; n < h[h_map[i]]->GetNbinsX(); ++n) {
				fprintf(tv_out,"%d\n",(int)h[h_map[i]]->GetBinContent(n));
			}
			fclose(tv_out);

			PRINTD("Opening calibration output file: %s\n",calbuf);
			tv_out = fopen(calbuf,"w");
			if(!tv_out) {
				printf("Could not open tv calibration output file '%s'\n.",calbuf);
				continue;
			}

			//fprintf(tv_out,
			//"#Offset\n%f\n#Slope\n%f\n\n",calib_offsets[i],calib_slopes[i]);
			if(i < 8) { // For LaBr
				fprintf(tv_out,"%.11f\n%.11f\n%.11f\n%.11f\n",
				    calib_offsets[i],calib_slopes[i],calib_quad[i],calib_cubic[i]);
			} else {
				fprintf(tv_out,"%f\n%f\n\n",calib_offsets[i],calib_slopes[i]);
			}
			fclose(tv_out);

		}
	}

	TTimer* t1, *t2;
	TFile* f;
	TTree* ntuple;
	Int_t first;
	TGMainFrame* frame;
	TGTab* tab;
	TRootEmbeddedCanvas* ec_l, *ec_h, *ec_3, *ec_4, *ec_5, *ec_6, *ec_7, *ec_8, *ec_9;
	TRootEmbeddedCanvas* ec[10];
	myTH1F* h[64]; 
	myTH1F* hc[48]; 
	TH2F* h2[16];
	TGCompositeFrame* labr_frame, *hpge_frame, *frame_3, *frame_4, 
	    *frame_5, *frame_6, *frame_7, *frame_8, *frame_9;
	TGCompositeFrame* cf[10];
	TRandom r;
	TGStatusBar* status;
	char* filename;

	/* Signals */
	//void StatusBarChanged(); //*SIGNAL*

	/* Slots */
	void RedrawStatusBar() {
		status->Draw();
	}


	/* Branches */
	TBranch* b_labr_h_e[8];
	TBranch* b_labr_l_e[8];
	TBranch* b_labr_t[8];
	TBranch* b_labrl_t[8];
	TBranch* b_labr_sis_e[4];
	TBranch* b_labr_sis_b[4];
	TBranch* b_hpge_e[4];
	TBranch* b_hpge_b[4];
	TBranch* b_hpge_pileup[4];
	TBranch* b_hpge_t[4];
	TBranch* b_hpgel_t[4];
	TBranch* b_labr_s[8];
	TBranch* b_labrl_s[8];
	TBranch* b_hpge_s[4];
	TBranch* b_hpge_crm[4];
	TBranch* b_hpgel_s[4];
	TBranch* b_trig_s[16];
	TBranch* b_trig_t[16];
	TBranch* b_trlo_sc[64];
	TBranch* b_tpat;
	TBranch* b_t0;
	TBranch* b_time_lo;
	TBranch* b_time_hi;
	/* Variables */
	UInt_t   labr_h_e[8];
	UInt_t   labr_l_e[8];
	UInt_t   labr_t[8];
	UInt_t   labrl_t[8];
	UInt_t    labr_s[8];
	UInt_t    labr_s_cur[8];
	UInt_t    labr_s_prev[8];
	UInt_t    labr_rate[8];
	UInt_t    labrl_s[8];
	UInt_t    labrl_s_cur[8];
	UInt_t    labrl_s_prev[8];
	UInt_t    labrl_rate[8];
	UInt_t   labr_sis_e[4];
	UInt_t   labr_sis_b[4];
	UInt_t    hpge_s[4];
	UInt_t    hpge_s_cur[4];
	UInt_t    hpge_s_prev[4];
	UInt_t    hpge_rate[4];
	UInt_t    hpgel_s[4];
	UInt_t    hpgel_s_cur[4];
	UInt_t    hpgel_s_prev[4];
	UInt_t    hpge_crm[4];
	UInt_t    hpge_crm_cur[4];
	UInt_t    hpge_crm_prev[4];
	UInt_t    hpge_crm_rate[4];
	Int_t    hpgel_rate[4];
	UInt_t    hpge_e[4];
	UInt_t    hpge_b[4];
	UInt_t   hpge_pileup[4];
	UInt_t   hpge_t[4];
	UInt_t   hpgel_t[4];
	UInt_t   tpat;
	UInt_t   t0;
	UInt_t    trig_s[16];
	UInt_t    trig_s_prev[16];
	UInt_t    trig_rate[16];
	UInt_t   trig_t[16];
	UInt_t    time_lo;
	UInt_t    time_hi;
	trlo_scaler trlo_sc[16];
	trlo_scaler64 trlo_sc_sum[16];
	trlo_scaler64 trlo_sc_start[16];
	float labr_h_e_calib[8];
	float labr_sis_e_calib[8];
	float labr_l_e_calib[8];
	float hpge_e_calib[4];
	float calib_slopes[32];
	float calib_offsets[32];
	float calib_cubic[32];
	float calib_quad[32];
	uint64_t start, stop;
	uint64_t ts_first, ts_last, timestamp_begin;
};

void ghost(char* filename, char* options = 0) {
	TApplication a;
	printf("Options = %s\n",options);
	new TreeReader(filename,options);
	a.Run();
	//gPad->AddExec("myexec","myexec()");
}
