#include <TCanvas.h>
#include <TTree.h>
#include <TFile.h>
#include <TBranch.h>
#include <TString.h>
#include <unistd.h>
#include <TH1F.h>
#include <TF1.h>
#include <TPad.h>
#include <TSystem.h>

#define USE_TRACES
#define N_CH 4
#define N_MAW	1022

enum DetectorAdc
{
	ADC_LABR = 1,
	ADC_NDET = 2,
	ADC_HPGE = 3
};

#define N_SAMPLES_MAX 10000
const unsigned int n_samples[] =
{
	0,
	150,
	500,
	N_SAMPLES_MAX
};

const unsigned int thresholds[] =
{
	10000,
	100,
	0,
	6000 /* 20000 at 100 ns risetime */
};

const double maw_max[] =
{
	1e6,
	1e5,
	1e5,
	1e5
};

TTree *t;
TCanvas *c;
TCanvas *cr;

TH1F* he[N_CH];
TH1F* ht[N_CH];
TH1F* hmawe[N_CH];
TH1F* he_trace[N_CH];

TH1F* hs[N_CH];
TH1F* hd[N_CH];
TH1F* hmaw[N_CH];
TH1F* hmaw_bl[N_CH];

unsigned int g_bytes_read = 0;

TBranch* b_v[N_CH];
TBranch* b_maw[N_CH];
TBranch* b_m[N_CH];
TBranch* b_b[N_CH];

unsigned int v[N_CH][N_SAMPLES_MAX];
unsigned int maw[N_CH][N_MAW];
unsigned int m[N_CH];
unsigned int b[N_CH];

void set_branch_addresses(enum DetectorAdc det);
void analyse_trace(unsigned int, unsigned int, enum DetectorAdc);

int stop = 0; /* 0: run through, 1: stop after each event */

void
analyse_hpge_traces(const char *filename, int _det, int _stop)
{
	TFile *f = new TFile(filename);
	if (f == NULL) {
		printf("Could not find file '%s'\n", filename);
		abort();
	}

	enum DetectorAdc det = (enum DetectorAdc)_det;
	stop = _stop;
	unsigned int i;
	unsigned int n;
	unsigned int ch;

	t = (TTree *) f->Get("h101");
	if (t == NULL) {
		printf("Could not find tree 'h101'\n");
		abort();
	}

	n = t->GetEntries();
	printf("%d entries.\n", n);

	c = new TCanvas("HPGe","HPGe");
	c->Divide(4,4);
	cr = new TCanvas("Results","Results");
	cr->Divide(4,4);

	for (i = 0; i < N_CH; ++i) {
		char buf[200];
		he[i] = new TH1F(TString::Format("e_%d",i),
		    "e", 1000, 0, 3000);
		ht[i] = new TH1F(TString::Format("tau_%d",i),
		    "tau", 500, 1e-4, 2e-4);
		hmawe[i] = new TH1F(TString::Format("mawe_%d",i),
		    "mawe", 4000, 0, maw_max[det]);
		he_trace[i] = new TH1F(TString::Format("integral_%d",i),
		    "integral", 4000, 0, 1e5);
		hs[i] = new TH1F("raw", "raw", n_samples[det], 0,
		    n_samples[det]);
		hs[i]->GetXaxis()->SetRange(0,n_samples[det]);
		hd[i] = new TH1F("smoothed", "smoothed",
		    n_samples[det],0,n_samples[det]);
		hd[i]->GetYaxis()->SetRange(-500,500);
		hd[i]->GetXaxis()->SetRange(1000,n_samples[det]);
		sprintf(buf, "maw %d", i+1);
		hmaw[i] = new TH1F(buf, buf, N_MAW, 0, N_MAW);
		sprintf(buf, "maw baseline %d", i+1);
		hmaw_bl[i] = new TH1F(buf, buf, N_MAW, 0, N_MAW);
		hmaw[i]->GetXaxis()->SetRange(0,N_MAW);
		hmaw_bl[i]->GetXaxis()->SetRange(0,N_MAW);
		hmaw_bl[i]->GetYaxis()->SetRangeUser(-5000,5000);
	}


	set_branch_addresses(det);

	for (i = 0; i < n; ++i) {
		for (ch = 0; ch < N_CH; ++ch) {
			analyse_trace(i, ch, det);
		}
	}

	for (i = 0; i < N_CH; ++i) {
		cr->cd(4*i+1);
		ht[i]->Draw();
		cr->cd(4*i+2);
		he[i]->Draw();
		cr->cd(4*i+3);
		hmawe[i]->Draw();
		cr->cd(4*i+4);
		he_trace[i]->Draw();
	}

	gPad->Update();
	gSystem->ProcessEvents();
}

void
set_branch_addresses(enum DetectorAdc det)
{
	unsigned int i;

	printf("det = %d\n",det);

	for (i = 0; i < N_CH; ++i) {
		char buf[200];
		unsigned int ch = (i * 4) + 1;

#ifdef USE_TRACES
		sprintf(buf, "ADC%dCH%dRAWv", det, ch);
		printf("Using %s\n", buf);
		t->SetBranchAddress(buf, v[i], &b_v[i]);
		sprintf(buf, "ADC%dCH%dMAWTv", det, ch);
		t->SetBranchAddress(buf, &maw[i], &b_maw[i]);
#endif
		sprintf(buf, "ADC%dCH%dMAXE1", det, ch);
		t->SetBranchAddress(buf, &b[i], &b_b[i]);
		sprintf(buf, "ADC%dCH%dMAXE2", det, ch);
		t->SetBranchAddress(buf, &m[i], &b_m[i]);
	}
}

#define N_BASELINE 30
#define N_AVG 20
#define N_GAP 20
#define T_THR -200

void
analyse_trace(unsigned int entry, unsigned int ch, enum DetectorAdc det)
{
	int e;
	int i;
	int j;

	int integrate_from;
	int integrate_to;

	int s[n_samples[det]];
	int d[n_samples[det]];
	int sum;
	int sum2;
	int integral;
	int bl;
	int t0;
	int x1, x2;
	double p[3];
	double *p_result;
	char buf[30];

	/*printf("<%d,%d> Analyse BEGIN\n", i, ch);*/

#ifdef USE_TRACES
	g_bytes_read += b_v[ch]->GetEntry(entry);
	g_bytes_read += b_maw[ch]->GetEntry(entry);
#endif
	g_bytes_read += b_m[ch]->GetEntry(entry);
	g_bytes_read += b_b[ch]->GetEntry(entry);

	e = m[ch] - b[ch];
	hmawe[ch]->Fill(e);

	if (e >= (int)thresholds[det]) {
		printf("Suitable trace %d in ch %d with energy value %d\n",
		    entry, ch, e);
	} else {
		printf("Skipping trace %d in ch %d (e = %d)\n", entry, ch, e);
		return;
	}

#ifdef USE_TRACES
	integral = 0;
	integrate_from = 60;
	integrate_to = 100;


	bl = 0;
	sum = 0;
	sum2 = 0;
	t0 = 0;
	printf("Smoothing...\n");
	for (i = 0; i < (int)n_samples[det] - N_AVG - N_GAP; ++i) {
		/* Accumulate baseline */
		if (i < N_BASELINE) { bl += v[ch][i]; }
		/* Devide baseline */
		if (i == N_BASELINE) { bl /= N_BASELINE; }
		/* Running average */
		if (i == 0) {
			for (j = 0; j < N_AVG; ++j) {
				sum += v[ch][j];
				sum2 += v[ch][j + N_GAP];
			}
		} else {
			sum -= v[ch][i];
			sum += v[ch][i + N_AVG];
			sum2 -= v[ch][i + N_GAP];
			sum2 += v[ch][i + N_GAP + N_AVG];
		}

		/* Integrate region */
		if (i > integrate_from && i <= integrate_to) {
			integral += v[ch][i];
		}

		/* New signals */
		if (i > N_BASELINE) {
			s[i] = sum/N_AVG - bl;
			d[i] = (sum-sum2)/N_AVG;
		} else {
			s[i] = 0;
			d[i] = 0;
		}

		if (t0 == 0 && d[i] < T_THR) {
			t0 = i;
		}
	}

	integral -= bl * (integrate_to - integrate_from);

	printf("Entry: %d, Integral = %d\n", entry, integral);

	for (i = 0; i < (int)n_samples[det]; ++i) {
		hs[ch]->SetBinContent(i, (int)v[ch][i]-bl);
	}

	for (i = 0; i < (int)n_samples[det]; ++i) {
		hd[ch]->SetBinContent(i, s[i]);
	}

	for (i = 0; i < N_MAW; ++i) {
		int val = maw[ch][i]-b[ch];
		hmaw[ch]->SetBinContent(i, val);
		hmaw_bl[ch]->SetBinContent(i, val);
	}

	/* Fit exponential */
	x1 = t0 + 2*N_AVG + N_GAP;
	x2 = 8000;

	/* Start values */
	p[0] = v[ch][x1]; 	/* Constant factor */
	p[1] = x1; 		/* Constant shift */
	p[2] = 0.002;		/* Tau factor */

	TF1* f = new TF1("myexp","[0] * exp(-[2] * (x - [1]))");
	f->SetParameters(p);

/*	c->cd(1);
	hs->Fit(f,"NR","",x1,x2);
	p_result = f->GetParameters();

	he[ch]->Fill(p_result[0]);
	ht[ch]->Fill(p_result[2]);*/

	he_trace[ch]->Fill(integral);

	if(stop == 1) {
		c->cd(4*ch+1);
		hs[ch]->Draw();
		c->cd(4*ch+2);
		hd[ch]->Draw();
		c->cd(4*ch+3);
		hmaw[ch]->Draw();
		c->cd(4*ch+4);
		hmaw_bl[ch]->Draw();
		gPad->Update();
		gSystem->ProcessEvents();
		getc(stdin);
	}

#endif
	/*printf("<%d,%d> Analyse END\n", i, ch);*/
}
