#ifndef PARSE_H
#define PARSE_H

#include <errno.h>
#include <stdlib.h>
#include <sys/queue.h>
#include <ctype.h>
#include <string.h>

struct Hist
{
	const char *name;
	int bins;
	double min;
	double max;
	TAILQ_ENTRY(Hist) next;
};

TAILQ_HEAD(HistList, Hist);
struct Tab
{
	const char *name;
	int nx;
	int ny;
	struct HistList *h;
	TAILQ_ENTRY(Tab) next;
};

enum Keyword
{
	KW_TAB,
	KW_H1,
	KW_H2,
	KW_THIS_IS_LAST
};

static const char kw[][48] =
{
	"tab",
	"h1",
	"h2"
};


TAILQ_HEAD(TabList, Tab);
struct TabList *g_tabs;

static int g_lines = 0;
static FILE *in;

struct TabList* parse_config(const char *);
int parse_tab(const char **);
void parse_tab_args(const char **, struct Tab*);
int parse_histogram(const char **, struct Tab*);
void swallow_whitespace(const char**);
void parse_stop(const char *);
int expect(const char *, const char **);
void expect_or_die(const char *, const char **);
const char *expect_str(const char **);
int expect_int(const char **);
enum Keyword expect_kw_any_of(enum Keyword[], size_t, const char **);
enum Keyword expect_keyword(enum Keyword, const char **);
void get_next_line(const char **);

int
expect(const char *what, const char **a_p)
{
	const char *p = *a_p;

	/*printf("expect %s\n", what);*/
	swallow_whitespace(&p);
	if (0 != strncmp(what, p, strlen(what))) {
		/*printf("Expected '%s' and got '%s'.\n",
		    what, p);*/
		return -1;
	}
	p += strlen(what);
	*a_p = p;
	return 0;
}

void
expect_or_die(const char *what, const char **a_p)
{
	int ret;
	ret = expect(what, a_p);
	if (ret == -1) {
		printf("Expected '%s' and got '%s' \n",
		    what, *a_p);
		printf("Aborting...\n");
		abort();
	}
	printf("expect_or_die('%s'), rest '%s'\n", what, *a_p);
}

enum Keyword
expect_kw_any_of(enum Keyword kw_list[], size_t len, const char **a_p)
{
	unsigned int i;
	enum Keyword ret;

	const char *p = *a_p;

	for(i = 0; i < len / sizeof(kw_list[0]); ++i) {
		ret = expect_keyword(kw_list[i], &p);
		if (ret == kw_list[i]) {
			*a_p = p;
			return ret;
		}
	}

	printf("Expected any of: ");
	for(i = 0; i < len / sizeof(kw_list[0]); ++i) {
		printf("%s ", kw[kw_list[i]]);
	}
	printf("but got %s\n", p);

	return (enum Keyword)-1;
}

enum Keyword
expect_keyword(enum Keyword a_kw, const char **a_p) {
	const char *p = *a_p;
	int ret;

	ret = expect(kw[a_kw], &p);
	if (ret == -1) return (enum Keyword)ret;

	*a_p = p;
	return a_kw;
}

const char *
expect_str(const char **a_p)
{
	const char *p = *a_p;
	char *str;
	size_t len = 0;

	expect_or_die("\"", &p);
	++(*a_p);
	while(isalnum(*p) || ' ' == *p) {
		++len;
		++p;
	}
	expect_or_die("\"", &p);
	if (len == 0) {
		printf("Expected str, but got '%s'\n", *a_p);
		return NULL;
	}

	str = strndup(*a_p, len);

	printf("Found string: %s\n", str);

	*a_p = p;
	return str;
}

int
expect_int(const char **a_p)
{
	const char *p = *a_p;
	int ret;
	size_t len = 0;
	char *str;

	swallow_whitespace(&p);
	*a_p = p;
	while(isdigit(*p)) {
		++p;
		++len;
	}
	if (len == 0) {
		printf("Expected int, but got '%s'\n", *a_p);
		abort();
	}

	str = strndup(*a_p, len);
	
	/*printf("Found string: %s with len %lu\n", str, len);*/
	
	ret = atoi(str);

	printf("Found int: %d\n", ret);

	*a_p = p;
	return ret;
}

void
parse_stop(const char *msg)
{
	printf("%d: Parse stopped: %s\n", g_lines, msg);
	abort();
}

struct TabList*
parse_config(const char *name)
{
	int ret;
	const char *p = NULL;

	in = fopen(name, "r");
	if (in == NULL) {
		perror("fopen");
		abort();
	}

	printf("Opened file %s.\n", name);
	swallow_whitespace(&p);

	g_tabs = (TabList *) malloc(sizeof(struct TabList));
	TAILQ_INIT(g_tabs);

	do {
		ret = parse_tab(&p);
	} while (ret == 1);

	return g_tabs;
}

int
parse_tab(const char **a_p)
{
	static int count = 0;
	int ret;
	const char *p = *a_p;

	struct Tab *t;
	t = (struct Tab *) malloc(sizeof(struct Tab));
	t->h = (struct HistList *) malloc(sizeof(struct HistList));
	TAILQ_INIT(t->h);

	if(p == NULL) return 0;

	expect_or_die(kw[KW_TAB], &p);	
	parse_tab_args(&p, t);

	expect_or_die("{", &p);
	for(;;) {
		ret = expect("}", &p);
		if (ret == 0) {
			 break;
		}
		ret = parse_histogram(&p, t);
	}

	count++;
	TAILQ_INSERT_TAIL(g_tabs, t, next);
	printf("Inserting Tab[%d] = {name = %s, nx = %d, ny = %d}\n",
	    count, t->name, t->nx, t->ny);

	*a_p = p;

	return 1;
}

void
parse_tab_args(const char **a_p, struct Tab *t)
{
	const char *p = *a_p;

	expect_or_die("(", &p);
	t->name = expect_str(&p);
	expect_or_die(",", &p);
	t->nx = expect_int(&p);
	expect_or_die(",", &p);
	t->ny = expect_int(&p);
	expect_or_die(")", &p);

	*a_p = p;
}

int
parse_histogram(const char **a_p, struct Tab *t)
{
	static int count = 0;
	const char *p = *a_p;

	enum Keyword kw_list[] = {
		KW_H1,
	       	KW_H2
	};

	int h_type;
	struct Hist *h;

	h_type = expect_kw_any_of(kw_list, sizeof(kw_list), &p);
	if (h_type == -1) {
		return -1;
	}

	h = (struct Hist *) malloc(sizeof(struct Hist));

	expect_or_die("(", &p);
	h->name = expect_str(&p);
	expect_or_die(",", &p);
	h->bins = expect_int(&p);
	expect_or_die(",", &p);
	h->min = expect_int(&p);
	expect_or_die(",", &p);
	h->max = expect_int(&p);
	expect_or_die(")", &p);

	count++;
	TAILQ_INSERT_TAIL(t->h, h, next);
	printf("Inserting Hist[%d] = {name = %s, bins = %d, min = %f, max = %f}"
	    "\n", count, h->name, h->bins, h->min, h->max);

	*a_p = p;

	return 1;
}

void
swallow_whitespace(const char **a_p)
{
	const char *p = *a_p;

	if (p == NULL || *p == '\0') {
		get_next_line(&p);
	}
	/*printf("swallow_whitespace, len = %lu\n", strlen(p));*/
	while(isspace(*p)) {
		printf("Found space '%c' in '%s'\n",*p, p);
		p++;
		if (p == NULL || *p == '\0') {
			get_next_line(&p);
			if (p == NULL) {
				return;
			}
		}
	}

	*a_p = p;
}

void
get_next_line(const char **a_p) {
	static char *last_line = NULL;
	ssize_t size;
	size_t len = 0;
	char *line = NULL;

	size = getline(&line, &len, in);
	printf("line '%s' with len = %lu, size = %lu\n", line, len, size);
	if (0 > size) {
		if (errno == 0) {
			*a_p = NULL;
			return;
		}
		perror("getline");
		parse_stop("Getline failed while whitespace");
	}
	free((void *)last_line);
	*a_p = line;
	last_line = line;

	printf("Read new line: %s\n", line);

	g_lines++;
}

#endif
